from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.devices.observer import Observer


class DeviceChannel(Device, Observer):
    """
    Base class for Device acting as a Channel of parent Device.
    This class serves only as a Template and must be inherited by target Class.
    Typical example is Motor axis, Interferometer channel or any Device which can control multiple sub Devices (Channels).
    Example of config file:

    Motor1_axis_0: &x !Device:kamzik3.devices.dummy.dummyMotor.DummyMotorChannel
        device: *Parent_Device
        channel: 0
        device_id: Motor_axis
    """

    def __init__(self, device, channel, device_id=None, config=None):
        """
        Base logic is to use subject_update method to observe parent Device.
        Parent Device then should call notify method to notify all observers about update.
        On Channel level, we should then filter for channel ID, which is UINT starting from 0.
        You are free to reimplement subject_update to Your needs.
        Usually Parent Device call notify(key, value) with parameters key=(channel_id, ATTR_NAME), value.
        Subject is always parent Device, called notify, since we can observe multiple Devices.
        :param Device device: parent Device to observe
        :param uint channel: channel id starting from 0
        :param str device_id: Device ID
        :param dict config: Device configuration
        """
        self.configured = False
        self.device = device
        self.channel = channel
        self.position_attribute_copy = None
        super().__init__(device_id, config)
        self.set_status(STATUS_CONNECTING)
        self.device.attach_observer(self)

    def _init_attributes(self):
        Device._init_attributes(self)

    def command(self, command, callback=None, with_token=False, returning=True):
        """
        Wrapper for Device command.
        Instead of executing command on this Device, execute it on parent Device.
        :param str command: Request to Device
        :param Callable callback: callback after Device replies to Command
        :param bool with_token: token
        :param returning: bool
        :return int: token
        """
        return self.device.command(command, callback, with_token, returning)

    def subject_update(self, key, value, subject):
        """
        This method should be overloaded in Target class.
        Here should go all the logic, that filters for channel id and does update of Channel Device Attribute.
        :param mixed key: subject name
        :param mixed value: subject value
        :param Subject subject: Subject
        """
        if key == ATTR_STATUS:
            if value in READY_DEVICE_STATUSES:
                self.handle_configuration()
            else:
                self.configured = False
                self.set_status(value)

    def poll_command(self, command, interval):
        """
        Wrapper for poll_command.
        Register polled command on Parent Device.
        :param str command: Request / Command
        :param int interval: interval in ms
        """
        self.device.poll_command(command, interval)

    def remove_poll_command(self, command, interval):
        """
        Wrapper for remove_poll_command.
        Unregister polled command on Parent Device.
        :param str command: Request / Command
        :param int interval: interval in ms
        """
        self.device.remove_poll_command(command, interval)

    def handle_configuration(self):
        raise NotImplementedError("Must be overloaded")

    def disconnect(self):
        self.stop_polling()
        self.configured = False
        self.device.detach_observer(self)
        self.set_status(STATUS_DISCONNECTED)

    # pylint: disable=unused-argument
    def reconnect(self, *args):
        self.device.attach_observer(self)
        self.handle_configuration()



import json
import logging
from threading import Thread

import numpy as np

from kamzik3.constants import MSG_JSON, MSG_ARRAY
from kamzik3.snippets.snippetsJson import JsonKamzikEncoder
from kamzik3.snippets.snippetsZmq import Publisher


class DevicePublisher(Publisher, Thread):
    def __init__(self, host, port, zmq_context=None):
        self.logger = logging.getLogger("Console.DeviceServer@{}:{}".format(host, port))
        Thread.__init__(self)
        Publisher.__init__(self, host, port, zmq_context)

    def _publish_message(self, header, message):
        try:
            if isinstance(message, np.ndarray):
                self.socket_publisher.send_multipart(
                    [
                        header.encode(),
                        MSG_ARRAY,
                        str(message.dtype.name).encode(),
                        json.dumps(message.shape).encode(),
                        message,
                    ],
                    copy=False,
                )
            else:
                json_message = json.dumps(
                    message, cls=JsonKamzikEncoder, ensure_ascii=True
                )
                self.socket_publisher.send_multipart(
                    [header.encode(), MSG_JSON, json_message.encode()], copy=False
                )
        except Exception as e:
            self.logger.error("{}, {}, {}".format(e, header, message))



from ctypes import CDLL

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device


class DeviceOsLibrary(Device):
    def __init__(self, os_library, device_id=None, config=None):
        if not hasattr(self, "os_library"):
            self.os_library = os_library
        if not hasattr(self, "library"):
            self.library = None
        super().__init__(device_id, config)
        self.connect()

    # Is abstract in Device and must be overridden
    def handle_configuration(self) -> None:
        pass

    def _init_attributes(self):
        Device._init_attributes(self)
        self.create_attribute(
            ATTR_LIBRARY, default_value=self.os_library, readonly=True
        )

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...

        :param args: connect attributes
        """
        self.connecting = True
        self.handle_configuration_event()
        if self.library is None:
            try:
                self.library = CDLL(self.os_library)
            except OSError as e:
                raise DeviceError(
                    "Error loading library {}. {}".format(self.os_library, e)
                )
        self.connected = True
        self.connecting = False
        self.set_status(STATUS_IDLE)



import time

import numpy as np
import serial
from pint import Quantity
from typing import Optional

import kamzik3
from kamzik3 import units, WriteException
from kamzik3.constants import *
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket


class DeviceThermoCube(DeviceSocket):
    """
    Chiller device.
    Documentation is here: https://svnsrv.desy.de/desy/cfel-cxi/doc/manuals/thermo/ThermoCube.pdf
    The ThermoCube 200/300/400 can store 8 bytes of transmission and can only handle up to 3
    commands per second.
    """

    terminator: Optional[bytes] = b"\r\n"
    push_commands_max = 1

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        # Check for error
        self.command(b"\xC8")
        # Get temperature readout
        self.command(b"\xC9")
        # Get temperature setpoint
        self.command(b"\xC1", callback=_finish_configuration)

    def _init_attributes(self):
        """
        Attributes;
            ATTR_FAULT_BYTE: True if ERROR occurred
            ATTR_TEMPERATURE: Temperature readout
            ATTR_SETPOINT: Temperature setpoint
        """
        DeviceSocket._init_attributes(self)
        self.create_attribute(
            ATTR_FAULT_BYTE,
            readonly=True,
            min_value=0,
            max_value=8,
            default_type=np.uint8,
        )
        self.create_attribute(
            ATTR_TEMPERATURE,
            readonly=True,
            min_value=-100,
            max_value=100,
            unit="degC",
            default_type=np.float,
            decimals=3,
        )
        self.create_attribute(
            ATTR_SETPOINT,
            min_value=-100,
            max_value=100,
            unit="degC",
            default_type=np.float,
            decimals=3,
            set_function=self.set_setpoint,
        )

    def handle_readout(self, readout_buffer):
        """
        Handle readout from Device.
        To get value from readout we have to convert HEX to DECIMAL.
        :param str readout_buffer: raw readout buffer from Device
        """
        command, output, callback, token = DeviceSocket.handle_readout(
            self, readout_buffer
        )
        if command == b"\xC9":
            temperature = self.hex_to_temperature(
                self.parse_temperature_readout(output)
            )
            self.set_raw_value(ATTR_TEMPERATURE, temperature.m)
        elif command == b"\xC1":
            setpoint = self.hex_to_temperature(self.parse_temperature_readout(output))
            self.set_raw_value(ATTR_SETPOINT, setpoint.m)
        elif command == b"\xC8":
            errorByte = int(output)
            self.set_raw_value(ATTR_FAULT_BYTE, errorByte)
            if errorByte == 0:
                self.set_status(STATUS_IDLE)
            else:
                self.set_status(STATUS_ERROR)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(
                self.device_id, [command, readout_buffer], token
            )

    def start_polling(self):
        DeviceSocket.start_polling(self)
        # Poll current temperature
        self.poll_command(b"\xC9", 500)
        # Poll temperature setpoint
        self.poll_command(b"\xC1", 5000)
        # Poll for error
        self.poll_command(b"\xC8", 500)

    @staticmethod
    def parse_temperature_readout(readout) -> str:
        """
        Return HEX representation of temperature
        :param str readout: Raw Device readout
        :return hex: temperature readout in HEX
        """
        low_byte, high_byte = readout.split(",")
        return "0x%x%x" % (int(high_byte), int(low_byte))

    @staticmethod
    def temperature_to_hex(temperature) -> str:
        """
        Convert temperature to HEX
        :param float temperature: Temeprature setpoint
        :return str: Setpoint in HEX
        """
        temperature_in_kelvin = units.Quantity(temperature, "celsius").to("fahrenheit")
        return "{:04X}".format(int(temperature_in_kelvin.m * 10))

    @staticmethod
    def hex_to_temperature(hex_temperature) -> Quantity:
        """
        Convert temp from HEX to Pint units Quantity
        :param str hex_temperature: Temperature in HEX
        :return Quantity: Temperature in Pint units Quantity
        """
        temperature_in_fahrenheit = int(hex_temperature, 16) / 10
        return units.Quantity(temperature_in_fahrenheit, "fahrenheit").to("celsius")

    def get_temperature(self):
        """
        Get temperature readout.
        """
        self.command(b"\xC9")

    def get_setpoint(self):
        """
        Get temperature setpoint value.
        """
        self.command(b"\xC1")

    def get_error(self):
        """
        Get error flag
        """
        self.command(b"\xC8")

    # FIXME: Can't we just remove the argument here? Check back when we have a type for the argument to create_attribute
    def set_setpoint(self, temperature, callback=None):
        """
        Set temperature setpoint.
        We need to send three commands.
        E1, Low data byte, High data byte.
        :param float temperature: temperature setpoint
        :param callback: not used
        """
        temperature_in_hex = self.temperature_to_hex(temperature)
        self.command(b"\xE1", returning=False)
        self.command(bytes.fromhex(temperature_in_hex[2:]), returning=False)
        self.command(
            bytes.fromhex(temperature_in_hex[:2]), returning=False, with_token=True
        )

    def send_command(self, commands):
        """
        We need to modify a bit send_command routine.
        :param list commands: list of commands to be send
        :return list: list of NOT sent commands
        """
        try:
            if self.connected:
                command = commands.pop(0)
                if command[3]:  # If command is returning
                    self.commands_buffer.append((command, time.time()))
                else:  # Command is not returning, simulate immediate execution
                    self.response_timestamp = time.time()
                self.push(command[0])
            else:
                self.handle_response_error("Device is not connected")
        except IndexError:
            self.handle_connection_error(
                "Device {} buffer error".format(self.device_id)
            )
        except (WriteException, serial.SerialException):
            self.handle_response_error("Device {} writing error".format(self.device_id))
        finally:
            # FIXME: Is it an error if exceptions get swallowed?
            return commands


# abstract method `found_terminator` is defined in DeviceSocket (base class of
# DeviceThermoCube). Since multiple inheritance is used here we better disable the
# pylint warning.
# pylint: disable=abstract-method
# see https://github.com/python/mypy/issues/9319 for mypy error
class DeviceThermoCubePort(DevicePort, DeviceThermoCube):  # type: ignore
    """
    Extend ETH controller to enable RS232 connection.
    """

    terminator: Optional[bytes] = None
    push_commands_max = 1

    # pylint: disable=super-init-not-called
    def __init__(
        self,
        port,
        baud_rate=9600,
        parity=serial.PARITY_NONE,
        stop_bits=serial.STOPBITS_ONE,
        byte_size=serial.EIGHTBITS,
        device_id=None,
        config=None,
    ):
        DevicePort.__init__(
            self, port, baud_rate, parity, stop_bits, byte_size, device_id, config
        )

    def handshake(self) -> bool:
        """
        Confirm connection to Device via RS232.
        :return bool: True if handshake successful
        """
        self.logger.info("Handshake initiated")
        self.push(b"\xC8")
        time.sleep(0.3)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            readout = ord(self.read_all())
            if readout >= 0:
                return True
        return False

    def collect_incoming_data(self, data: bytes) -> None:
        """
        Modify collect_incoming_data method to match Device's communication protocol specification.
        :param bytes data: Device response bytes
        """
        # Get all bytes as separate string
        for c in data:
            self.buffer.append(str(c))

        cmd_query = self.commands_buffer[0][0][0]
        # Commands that queries temperature consist of two bytes
        if cmd_query in (b"\xC9", b"\xC1") and len(self.buffer) != 2:
            return
        # Create final string by joining buffer content with comma
        self.buffer = [",".join(self.buffer)]
        # Invoke found_terminator even though there is now terminator
        # We just want to continue in normal controller workflow and process command
        self.found_terminator()



import logging
import time
from logging import Logger
from logging.handlers import TimedRotatingFileHandler
from threading import Thread
from time import sleep
from typing import ValuesView

import numpy as np

from kamzik3 import DeviceError, DeviceUnknownError
from kamzik3.constants import *
from kamzik3.devices.attribute import Attribute
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetLogging import set_rotating_file_handler, set_file_handler
from kamzik3.snippets.snippetsTimer import PreciseCallbackTimer


class AttributeLogger(Device):
    """
    AttributeLogger is a parent Class for all derived Logger Classes.

    Logger itself is a Device.
    Logfile is rotating every day at 24:00.
    This Class is logging only Device's Attributes.
    Log entry is written every 'interval' tick.
    Logging starts automatically from the Session.

    Example of config file:
    session: !Device:kamzik3.devices.deviceSession.DeviceSession
        config:
            attributes:
              !!python/tuple [Log directory, Value]: ./server_log
            logged_attributes:
              # Device_Id: Device's Attribute
              # str: str | list
              X: Position
              Y: Position
              Z: Position
              Device: [Group, Attribute]
            # ... Other Session config attributes
    """

    # Logging interval in ms
    interval = 1e3
    # Value and separator separator
    separator = ";"
    logged_attributes = None
    attribute_logger = None
    attribute_logger_handler = None
    # User defined header, which is added before generated one
    preset_header = ""
    # Timer ticking at 'interval' frequency
    logger_ticker = None
    # Offset before start logging (10 seconds)
    start_offset = 10

    def __init__(self, log_file_name, device_id=None, config=None):
        # Check if logged_attributes were already set
        # (in case of Disconnected and Connected Device)
        if self.logged_attributes is None:
            self.logged_attributes = []
        # Holds actively logged attributes
        # This list is populated with Attribute's header
        self.active_log_attributes = []
        self.log_file_name = log_file_name
        # Define formatter for log entry
        self.log_formatter = logging.Formatter(
            self.separator.join(["%(asctime)s", "%(created)s", "%(message)s"]),
            datefmt="%Y-%m-%d %H:%M:%S",
        )
        Device.__init__(self, device_id, config)
        self.connect()

    def connect(self, *args):
        self.handle_connect_event()
        self.set_status(STATUS_IDLE)

    def _init_attributes(self):
        """
        Init Device's attributes.

        ATTR_HEADER: Header of the current logfile
        ATTR_INTERVAL: Logging interval in ms
        ATTR_LAST_LOG_LINE: Last line written in logfile
        ATTR_LOGFILE: Path to the logfile, might be absolute or relative
        ATTR_LOGGING: Start / Stop logging toggle
        """
        super()._init_attributes()
        self.create_attribute(ATTR_HEADER, readonly=True)
        self.create_attribute(
            ATTR_INTERVAL,
            default_value=self.interval,
            min_value=100,
            max_value=3600 * 1e3,
            unit="ms",
            default_type=np.uint32,
            set_function=self._set_interval,
        )
        self.create_attribute(ATTR_LAST_LOG_LINE, readonly=True)
        self.create_attribute(
            ATTR_LOGFILE, default_value=self.log_file_name, readonly=True
        )
        self.create_attribute(ATTR_LOGGING, default_type=bool, set_function=self.start)

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def _set_interval(self, interval):
        """
        Set tick interval.

        :param int interval: interval in ms
        :return:
        """
        self.interval = interval
        if self.is_status(STATUS_BUSY):
            # If logging is running, stop it and start again
            self.stop()
            self.start()

    def _get_logged_attributes(self) -> ValuesView:
        """
        Collect all logged attributes.

        If a new attribute is discovered, it generates and writes a header.

        :return: ValuesView
        """
        logged_values = dict.fromkeys(self.active_log_attributes, str(np.NaN))
        generate_header = False
        if self.logged_attributes is not None:
            for device_id, attribute in self.logged_attributes:
                try:
                    device = self.session.get_device(device_id)
                    device_attribute = device.get_attribute(attribute)

                    if device_attribute is None:
                        logged_value = None
                        logged_type = ""
                        logged_unit = ""
                    else:
                        logged_value = device_attribute[VALUE]
                        logged_type = device_attribute[TYPE]
                        logged_unit = device_attribute[UNIT]
                    if logged_type == "":
                        logged_type = "str"
                    if logged_unit == "":
                        logged_unit = "None"
                    header_lookup = (
                        f"{device_id} {attribute} ({logged_type}, {logged_unit})"
                    )
                    if header_lookup not in self.active_log_attributes:
                        self.active_log_attributes.append(header_lookup)
                        generate_header = True
                    if logged_value is None:
                        logged_value = np.NaN
                    logged_values[header_lookup] = str(logged_value)

                except DeviceUnknownError:
                    continue
        if generate_header:
            self.write_header()
        return logged_values.values()

    def set_log_file_name(self, log_file_name) -> Logger:
        """
        Create a new logfile with the defined name.

        The logging level INFO is used for Attribute logging.

        :param filepath log_file_name: Filepath to the file
        :return: Logger
        """
        self.set_raw_value(ATTR_LOGFILE, log_file_name)
        attribute_logger = logging.getLogger(f"AttributeLogger.{self.device_id}")
        if self.attribute_logger_handler is not None:
            attribute_logger.removeHandler(self.attribute_logger_handler)

        if self.config.get("rotating", True):
            # Setup rotating log, if 'rotating' attribute is set
            self.attribute_logger_handler = set_rotating_file_handler(
                attribute_logger, self.log_file_name, self.log_formatter
            )

            def do_rollover():
                """
                Perform the rollover.

                We need to write the header at the beginning of each rotated file.
                """
                TimedRotatingFileHandler.doRollover(self.attribute_logger_handler)
                self.write_header()

            self.attribute_logger_handler.doRollover = do_rollover
        else:
            # In this case we have only one log file
            self.attribute_logger_handler = set_file_handler(
                attribute_logger, self.log_file_name, self.log_formatter
            )
        attribute_logger.setLevel(logging.INFO)
        return attribute_logger

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.attribute_logger = self.set_log_file_name(self.log_file_name)
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        _finish_configuration()

    def handle_interval_timeout(self):
        """
        Callback for interval tick.

        In this method we write actual log entry.
        """
        logged_values = self._get_logged_attributes()
        if len(self.active_log_attributes) == 0:
            # If there is nothing to log, just return and don't write anything
            return
        last_log_line = self.separator.join(logged_values)
        self.set_raw_value(ATTR_LAST_LOG_LINE, last_log_line)
        if self.attribute_logger is not None:
            self.attribute_logger.info(last_log_line)

    def add_logged_attribute(self, device_id, attribute):
        """
        Add logged attribute.

        :param str device_id: Device ID
        :param Union[list, str, tuple] attribute: Device's Attribute
        """
        attribute = Attribute.list_attribute(attribute)
        item = (device_id, attribute)

        if item not in self.logged_attributes:
            self.logged_attributes.append(item)

    def remove_logged_attribute(self, device_id, attribute):
        """
        Remove logged Attribute.

        :param str device_id: Device ID
        :param Union[list, str, tuple] attribute: Device's Attribute
        """
        attribute = Attribute.list_attribute(attribute)
        self.logged_attributes.remove((device_id, attribute))

    def generate_header(self) -> str:
        """
        Generate the header for the log file.

        :return: str
        """
        header_line = [
            "Datetime ({}, {})".format("str", "None"),
            "Timestamp ({}, {})".format(np.dtype("float").name, "sec"),
        ]

        header_line += self.active_log_attributes
        return "# " + self.separator.join(header_line).replace("'", "").replace(
            ", ", ","
        )

    def write_line(self, line):
        """
        Write one line into log file.

        :param str line: formatted log line entry
        """
        self.attribute_logger_handler.flush()
        # Do we want to specify an encoding? Currently, it's system dependent
        # pylint: disable=unspecified-encoding
        with open(self.log_file_name, "a+") as fp:
            fp.write(line)
            fp.write("\n")

    def write_header(self):
        """
        Write header to the log file.

        New header is created from preset_header + generated_header
        """
        header = self.preset_header + self.generate_header()
        self.set_raw_value(ATTR_HEADER, header)
        self.write_line(header)

    def start(self, flag=True):
        """
        Start logging.

        By default Timer is started ticking at 'interval' frequency.
        Override this method if You need different start of the logger.
        If flag is True start logger, stop otherwise.

        :param bool flag: Toggle flag
        """
        if flag:
            if self.logger_ticker is not None and self.is_status(STATUS_BUSY):
                self.logger_ticker.stop()
                raise DeviceError("Logging is already running")
            self.logger_ticker = PreciseCallbackTimer(
                self.interval, self.handle_interval_timeout, with_correction=True
            )
            self.set_status(STATUS_BUSY)

            def _start():
                sleep(self.start_offset)
                self.logger_ticker.start()

            Thread(target=_start).start()
        else:
            self.stop()

    def stop(self):
        """Stop logging."""
        try:
            self.logger_ticker.stop()
            self.set_status(STATUS_IDLE)
        except TypeError:
            pass

    def close(self):
        self.stop()
        super().close()

    def disconnect(self):
        # Remove all handlers
        self.stop()
        try:
            for handler in self.attribute_logger.handlers[:]:
                handler.close()
        except AttributeError:
            # No attribute_logger in which case just continue
            pass
        self.logged_attributes = None
        self.attribute_logger = None
        self.attribute_logger_handler = None
        self.logged_attributes = []
        self.log_formatter = None
        return super().disconnect()


class AttributeLoggerTriggered(AttributeLogger):
    """
    Logger is used for Macros.

    Instead of periodically adding a new line, we log only when externally triggered.
    """

    def start(self, flag=True):
        """
        Generate the header and wait for external triggers.

        :param bool flag: Toggle flag
        """
        if flag:
            if self.is_status(STATUS_BUSY):
                raise DeviceError("Logging is already running")

            for device_id, attribute in self.logged_attributes:
                try:
                    device = self.session.get_device(device_id)
                    device_attribute = device.get_attribute(attribute)
                    logged_type = device_attribute[TYPE]
                    logged_unit = device_attribute[UNIT]
                    if logged_type == "":
                        logged_type = "str"
                    if logged_unit == "":
                        logged_unit = "None"
                    header_lookup = (
                        f"{device_id} {attribute} ({logged_type}, {logged_unit})"
                    )
                    if header_lookup not in self.active_log_attributes:
                        self.active_log_attributes.append(header_lookup)
                except DeviceUnknownError:
                    continue

            self.write_header()
            self.set_status(STATUS_BUSY)
        else:
            self.stop()

    def trigger(self):
        """
        Get a external trigger.

        This method is a wrapper for handle_interval_timeout of Parent AttributeLogger.
        """
        self.handle_interval_timeout()

    def stop(self):
        """Stop logging."""
        try:
            self.set_status(STATUS_IDLE)
        except TypeError:
            pass



from kamzik3.constants import *
from kamzik3.devices.general.deviceShutter import DeviceShutter
from kamzik3.snippets.snippetsDecorators import expose_method


class DummyShutter(DeviceShutter):
    """
    Implementation of Dummy Shutter, used for testing.
    """

    def _open_shutter(self):
        self.set_status(STATUS_IDLE)
        self.clear_open_delay()
        self.set_value(ATTR_OPENED, True)
        self.set_value(ATTR_OPENED_AT, 100)

    def _close_shutter(self):
        self.clear_close_delay()
        self.set_status(STATUS_IDLE)
        self.set_value(ATTR_OPENED, False)
        self.set_value(ATTR_OPENED_AT, 0)

    @expose_method()
    def is_opened(self):
        return self.get_value(ATTR_OPENED)



""" Example of yaml configuration
Smaract0: &Smaract0 !Device:kamzik3.devices.deviceSmaractMcs1.DeviceSmaractMcs1
    device_id: Smaract0
    host: 192.168.83.91
Smaract0_axis_0: &Smaract0_axis_0 !Device:kamzik3.devices.deviceSmaractMcs1.DeviceSmaractMcs1Channel
    device: *Smaract0
    channel: 0
    device_id: MotorAxis0
    config:
      attributes:
        !!python/tuple [Sensor type, Value]: 1
        !!python/tuple [Position +limit, Value]: 15mm
        !!python/tuple [Position -limit, Value]: -15mm
        !!python/tuple [Position, Tolerance]: [10nm, 10nm]
"""
import re
import time

import numpy as np
import serial
from bidict import frozenbidict
from pint.errors import UndefinedUnitError

import kamzik3
from kamzik3 import DeviceUnitError
from kamzik3.constants import *
from kamzik3.devices.deviceChannel import DeviceChannel
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

from kamzik3.constants import PREVIOUS


class DeviceSmaractMcs1(DeviceSocket):
    # push_buffer_size = 50
    response_timeout = 3000
    push_commands_max = 1
    terminator = b"\n"
    handtool_modes = frozenbidict({"Disabled": 0, "Enabled": 1, "Display only": 2})
    sensor_modes = frozenbidict({"Disabled": 0, "Enabled": 1, "Power safe": 2})
    channel_command_pattern = re.compile(":([A-Z]+)([-0-9]+),(.*)")
    linear_sensor_type = [None, 1, 5, 6, 9, 18, 21, 24, 32, 38, 40, 41, 42, 43, 44]
    sensor_types = [
        1,
        2,
        4,
        5,
        6,
        7,
        8,
        9,
        10,
        11,
        12,
        13,
        14,
        15,
        16,
        17,
        18,
        19,
        20,
        21,
        23,
        24,
        32,
        37,
        38,
        39,
        40,
        41,
        42,
        43,
        44,
        48,
        49,
        50,
        51,
        53,
        54,
    ]
    safe_direction_modes = frozenbidict({"Forward": 0, "Backward": 1})

    def __init__(self, host, port=5000, device_id=None, config=None):
        super().__init__(host, port, device_id, config)

    def _init_attributes(self):
        super()._init_attributes()
        self.create_attribute(ATTR_STATE, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_SERIAL_NUMBER, readonly=True)
        self.create_attribute(ATTR_CHANNELS, readonly=True, default_type=np.uint16)
        self.create_attribute(
            ATTR_HAND_TOOL_MODE, readonly=True, default_type=np.uint16
        )
        self.create_attribute(
            ATTR_HAND_TOOL_MODE,
            default_value="Display only",
            default_type=self.handtool_modes.keys(),
            set_function=self.set_hand_tool_mode,
        )
        self.create_attribute(
            ATTR_SENSOR_MODE,
            default_type=self.sensor_modes.keys(),
            set_function=self.set_sensor_mode,
        )
        self.create_attribute(
            ATTR_SENSOR_PRESENT, default_value=True, readonly=True, default_type=np.bool
        )

    def valid_command_format(self, command):
        return command is not None and command[-1] == "\n"

    def handle_configuration(self):
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        # Set communication to synchronous
        self.command(":SCM0\n")
        # Set hand-tool mode to just display positions
        self.set_hand_tool_mode("Display only")
        # Get Interface version
        self.command(":GSI\n")
        # Get number of channels
        self.command(":GNC\n")
        # Get Interface version
        self.command(":GSE\n", callback=_finish_configuration)

    def handle_readout(self, readout_buffer):
        command, output, callback, token = super().handle_readout(readout_buffer)

        if output[:2] == ":E":
            channel, error_number = [int(v) for v in output[2:].split(",")]
            if error_number > 0:
                if error_number == 129:
                    self.notify((channel, ATTR_SENSOR_MODE), "Disabled")
                    self.notify((channel, ATTR_SENSOR_PRESENT), False)
                # elif error_number == 140:
                #     self.set_offsetted_value(ATTR_SENSOR_MODE, "Disabled")
                #     for channel in range(self[ATTR_CHANNELS][VALUE]):
                #         self.notify((channel, ATTR_SENSOR_MODE), "Disabled")
                self.handle_command_error(command, output)
        elif command == ":GSI\n":
            self.set_offsetted_value(ATTR_SERIAL_NUMBER, output.strip(":IN"))
        elif command == ":GNC\n":
            channels = int(output.strip(":N"))
            self.set_offsetted_value(ATTR_CHANNELS, channels)
        elif command == ":GSE\n":
            mode = self.sensor_modes.inv[int(output.strip(":SE"))]
            self.set_offsetted_value(ATTR_SENSOR_MODE, mode)
            for channel in range(self[ATTR_CHANNELS][VALUE]):
                self.notify((channel, ATTR_SENSOR_MODE), mode)
        elif output[:3] == ":BR":
            pass
        else:
            find = self.channel_command_pattern.search(output)
            if find is not None and find != -1:
                attribute = None
                cmd, channel, value = find.group(1), int(find.group(2)), find.group(3)
                if cmd == "S":
                    attribute, value = ATTR_STATE, int(value)
                elif cmd == "P":
                    attribute, value = ATTR_POSITION, int(value)
                elif cmd == "A":
                    position, revolutions = value.split(",")
                    self.notify((channel, ATTR_REVOLUTIONS), int(revolutions))
                    attribute, value = ATTR_POSITION, int(position)
                elif cmd == "ST":
                    attribute, value = ATTR_SENSOR_TYPE, int(value)
                elif cmd == "PPK":
                    attribute, value = ATTR_REFERENCED, bool(int(value))
                elif cmd == "SD":
                    safe_direction = self.safe_direction_modes.inv[int(value)]
                    attribute, value = ATTR_SAFE_DIRECTION, safe_direction
                elif cmd == "CLA":
                    attribute, value = ATTR_ACCELERATION, int(value)
                elif cmd == "CLS":
                    attribute, value = ATTR_VELOCITY, int(value)
                elif cmd == "AL":
                    min_position, min_revolution, max_position, max_revolution = [
                        int(i) for i in value.split(",")
                    ]
                    attribute, value = ATTR_POSITION_LOWER_LIMIT, (
                        min_position,
                        min_revolution,
                    )
                    self.notify(
                        (channel, ATTR_POSITION_UPPER_LIMIT),
                        (max_position, max_revolution),
                    )
                elif cmd == "PL":
                    min_position, max_position = value.split(",")
                    min_position, max_position = int(min_position), int(max_position)
                    attribute, value = ATTR_POSITION_LOWER_LIMIT, int(min_position)
                    self.notify((channel, ATTR_POSITION_UPPER_LIMIT), int(max_position))
                if attribute:
                    self.notify((channel, attribute), value)

        if callback is not None:
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            kamzik3.session.publisher.push_message(
                self.device_id, [command, readout_buffer], token
            )
            for observer in self._observers[:]:
                if isinstance(observer, DeviceSmaractMcs1Channel):
                    kamzik3.session.publisher.push_message(
                        observer.device_id, [command, readout_buffer], token
                    )

    def set_hand_tool_mode(self, value):
        self.logger.info("Set hand tool mode to {}".format(value))
        int_mode = self.handtool_modes[value]
        self.command(":SHE{}\n".format(int_mode))

    def set_sensor_mode(self, value):
        self.logger.info("Set sensor mode to {}".format(value))
        for channel in range(self[ATTR_CHANNELS][VALUE]):
            self.notify((channel, ATTR_SENSOR_MODE), value)

        value = self.sensor_modes[value]
        self.command(":SSE{}\n".format(value))


# abstract methods `collect_incoming_data` and `found_terminator` are defined in
# DeviceSocket (one of the base classes of DeviceSmaractMcs1). Since multiple
# inheritance is used here we better disable the pylint warning.
# pylint: disable=abstract-method
# see https://github.com/python/mypy/issues/9319 for mypy error
class DeviceSmaractMcs1Port(DevicePort, DeviceSmaractMcs1):  # type: ignore
    push_buffer_size = DeviceSmaractMcs1.push_buffer_size
    response_timeout = DeviceSmaractMcs1.response_timeout
    terminator = DeviceSmaractMcs1.terminator

    def __init__(
        self,
        port,
        baud_rate=9600,
        parity=serial.PARITY_NONE,
        stop_bits=serial.STOPBITS_ONE,
        byte_size=serial.EIGHTBITS,
        device_id=None,
        config=None,
    ):
        super().__init__(
            port, baud_rate, parity, stop_bits, byte_size, device_id, config
        )

    def _init_attributes(self):
        super()._init_attributes()
        self.create_attribute(
            ATTR_BAUDRATE,
            default_value=self.baud_rate,
            readonly=False,
            default_type=self.BAUDRATES,
            set_function=self.set_baudrate,
        )

    def handshake(self):
        self.logger.info("Handshake initiated")
        self.push(b":SCM0\n")
        time.sleep(0.05)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            if self.read_all().strip() == ":E-1,0":
                return True
        return False

    def set_baudrate(self, value):
        self.logger.info("Set baud-rate to {}".format(value))
        self.command(":CB{}\n".format(int(value)))


class DeviceSmaractMcs1Channel(DeviceChannel):
    sensor_types = DeviceSmaractMcs1.sensor_types
    linear_sensor_type = DeviceSmaractMcs1.linear_sensor_type
    sensor_modes = DeviceSmaractMcs1.sensor_modes
    safe_direction_modes = DeviceSmaractMcs1.safe_direction_modes

    def _init_attributes(self):
        super()._init_attributes()
        self.create_attribute(ATTR_STATE, readonly=True, default_type=np.uint16)
        self.create_attribute(ATTR_MOVING, readonly=True, default_type=np.bool)
        self.create_attribute(
            ATTR_POSITION,
            default_type=np.int64,
            set_function=self._move_absolute,
            set_value_when_set_function=False,
        )
        self.create_attribute(ATTR_REVOLUTIONS, readonly=True, default_type=np.int64)
        self.create_attribute(ATTR_CALIBRATING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_REFERENCING, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_CALIBRATED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_REFERENCED, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_MOVEMENT_LOCKED, readonly=True, default_type=np.bool)
        self.create_attribute(
            ATTR_SENSOR_MODE,
            default_value=self.device.get_value(ATTR_SENSOR_MODE),
            readonly=True,
            default_type=self.sensor_modes.keys(),
        )
        self.create_attribute(
            ATTR_SENSOR_TYPE,
            default_type=self.sensor_types,
            set_function=self.set_sensor_type,
        )
        self.create_attribute(ATTR_SENSOR_PRESENT, readonly=True, default_type=np.bool)
        self.create_attribute(ATTR_ENCODER_BASE_UNIT, readonly=True)
        self.create_attribute(
            ATTR_ENCODER_RESOLUTION, readonly=True, default_type=np.int8
        )
        self.create_attribute(
            ATTR_POSITION_LOWER_LIMIT,
            default_type=np.int64,
            set_function=self.set_lower_position_limit,
        )
        self.create_attribute(
            ATTR_POSITION_UPPER_LIMIT,
            default_type=np.int64,
            set_function=self.set_upper_position_limit,
        )
        self.create_attribute(
            ATTR_STEP_FREQUENCY,
            default_value=1000,
            default_type=np.uint32,
            min_value=1,
            max_value=18.5e3,
            unit="Hz",
            set_function=self.set_step_frequency,
        )
        self.create_attribute(
            ATTR_STEP_AMPLITUDE,
            default_value=1000,
            default_type=np.uint32,
            min_value=1,
            max_value=4095,
            set_function=self.set_step_amplitude,
        )
        self.create_attribute(
            ATTR_ACCELERATION,
            default_type=np.int64,
            min_value=0,
            max_value=100e9,
            set_function=self.set_acceleration,
        )
        self.create_attribute(
            ATTR_VELOCITY,
            default_type=np.int64,
            min_value=0,
            max_value=1e8,
            set_function=self.set_velocity,
        )
        self.create_attribute(
            ATTR_SAFE_DIRECTION,
            default_type=self.safe_direction_modes.keys(),
            set_function=self.set_safe_direction,
        )
        self.create_attribute(
            ATTR_HOLDING_TIME,
            default_value=0,
            default_type=np.uint16,
            min_value=0,
            max_value=60e3,
            unit="ms",
        )
        self.create_attribute(ATTR_CLOSE_LOOP_UNIT, display=False)
        self.create_attribute(
            ATTR_OPEN_LOOP_FACTOR,
            default_value=[1, 1],
            default_type=TYPE_LIST,
            display=False,
        )
        self.create_attribute(
            ATTR_MOVE_METHOD,
            default_value=CLOSE_LOOP,
            default_type=[OPEN_LOOP, CLOSE_LOOP],
            set_function=self.set_move_method,
        )

    def handle_configuration(self):
        if self.configured:
            return

        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.set_value(ATTR_SENSOR_MODE, self.device.get_value(ATTR_SENSOR_MODE))
            self.set_value(
                ATTR_SENSOR_PRESENT, self.device.get_value(ATTR_SENSOR_PRESENT)
            )
            self.get_position_limit_range()
            self.start_polling()
            self.configured = True
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        self.connected = True
        self.get_sensor_type()
        self.get_acceleration()
        self.get_safe_direction()
        self.command(":GCLS{}\n".format(self.channel), callback=_finish_configuration)

    def subject_update(self, key, value, subject):
        super().subject_update(key, value, subject)
        if self.connected and isinstance(key, tuple) and key[0] == self.channel:
            attribute = key[1]
            if attribute == ATTR_STATE and self.get_value(ATTR_STATE) != value:
                if value in (0, 3):
                    self.set_status(STATUS_IDLE)
                    self.set_value(ATTR_CALIBRATING, False)
                    self.set_value(ATTR_REFERENCING, False)
                    self.set_value(ATTR_MOVING, False)
                else:
                    self.set_status(STATUS_BUSY)
                    if value in (1, 2, 4, 6, 7):
                        self.set_value(ATTR_MOVING, True)

                        if value == 6:
                            self.set_value(ATTR_CALIBRATING, True)
                        elif value == 7:
                            self.set_value(ATTR_REFERENCING, True)
                        elif value == 9:
                            self.set_value(ATTR_MOVEMENT_LOCKED, True)
            elif attribute == ATTR_SENSOR_TYPE:
                self._set_position_unit(value)
                self._set_position_polling(value)
                self.get_referenced()
                self.get_position_limit_range(value)
            elif attribute == ATTR_SENSOR_PRESENT:
                if not value:
                    self.remove_poll_command(":GP{}\n".format(self.channel), 250)
                    self.remove_poll_command(":GA{}\n".format(self.channel), 250)
                else:
                    self._set_position_polling(self[ATTR_SENSOR_TYPE][VALUE])
            elif attribute == ATTR_SENSOR_MODE:
                if value == "Disabled":
                    self.remove_poll_command(":GP{}\n".format(self.channel), 250)
                    self.remove_poll_command(":GA{}\n".format(self.channel), 250)
                else:
                    self._set_position_polling(self[ATTR_SENSOR_TYPE][VALUE])
            elif attribute == ATTR_POSITION and self[ATTR_REVOLUTIONS][VALUE] != 0:
                revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
                value += self[ATTR_REVOLUTIONS][VALUE] * revolution
            elif (
                attribute == ATTR_POSITION_LOWER_LIMIT
                and self[ATTR_SENSOR_TYPE][VALUE] not in self.linear_sensor_type
            ):
                revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
                min_position, min_revolutions = value
                value = min_position + min_revolutions * revolution
                # self.set_attribute([ATTR_POSITION, MIN], value)
            elif (
                attribute == ATTR_POSITION_UPPER_LIMIT
                and self[ATTR_SENSOR_TYPE][VALUE] not in self.linear_sensor_type
            ):
                revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
                max_position, max_revolutions = value
                value = max_position + max_revolutions * revolution
                # self.set_attribute([ATTR_POSITION, MAX], value)
            self.set_value(ATTR_LATENCY, self.device.get_value(ATTR_LATENCY))
            self.set_offsetted_value(attribute, value)

    def _set_position_unit(self, sensor_type):
        if sensor_type is None:
            return

        prefix = {-12: "p", -9: "n", -6: "u", -3: "m"}
        if sensor_type in self.linear_sensor_type:
            resolution = -9
            unit = "m"
        else:
            resolution = -6
            unit = "deg"
        base_position_unit = prefix.get(resolution, "") + unit
        self.set_attribute((ATTR_ENCODER_RESOLUTION, VALUE), resolution)
        self.set_attribute((ATTR_ENCODER_BASE_UNIT, VALUE), unit)
        self.set_attribute((ATTR_POSITION, UNIT), base_position_unit)
        self.set_attribute((ATTR_POSITION_UPPER_LIMIT, UNIT), base_position_unit)
        self.set_attribute((ATTR_POSITION_LOWER_LIMIT, UNIT), base_position_unit)
        self.set_attribute((ATTR_ACCELERATION, UNIT), "{}/s".format(base_position_unit))
        self.set_attribute((ATTR_VELOCITY, UNIT), "{}/s".format(base_position_unit))

    def _set_position_polling(self, sensor_type):
        if sensor_type is None:
            return
        elif sensor_type in self.linear_sensor_type:
            self.poll_command(":GP{}\n".format(self.channel), 250)
            self.set_value(ATTR_REVOLUTIONS, 0)
        else:
            self.poll_command(":GA{}\n".format(self.channel), 250)

    def start_polling(self):
        super().start_polling()
        self.poll_command(":GS{}\n".format(self.channel), 250)

    def stop_polling(self):
        self.remove_poll_command(":GS{}\n".format(self.channel), 250)
        self.remove_poll_command(":GP{}\n".format(self.channel), 250)
        self.remove_poll_command(":GA{}\n".format(self.channel), 250)

    def _move_absolute(self, value, callback=None):
        self.logger.info("Move to position {}".format(value))
        holding_time = self.get_value(ATTR_HOLDING_TIME)
        if self[ATTR_SENSOR_TYPE][VALUE] in self.linear_sensor_type:
            return self.command(
                ":MPA{},{},{}\n".format(self.channel, int(value), holding_time),
                callback,
                with_token=True,
            )
        else:
            revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
            revolutions = int(value // revolution)
            angular_position = int(value - revolutions * revolution)
            return self.command(
                ":MAA{},{},{},{}\n".format(
                    self.channel, angular_position, revolutions, holding_time
                ),
                callback,
                with_token=True,
            )

    def _move_relative(self, value, callback=None):
        self.logger.info("Move by relative position {}".format(value))
        holding_time = self.get_value(ATTR_HOLDING_TIME)
        if self[ATTR_SENSOR_TYPE][VALUE] in self.linear_sensor_type:
            return self.command(
                ":MPR{},{},{}\n".format(self.channel, int(value), holding_time),
                callback,
                with_token=True,
            )
        else:
            revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
            revolutions = int(value // revolution)
            angular_position = int(value - revolutions * revolution)
            return self.command(
                ":MAR{},{},{},{}\n".format(
                    self.channel, angular_position, revolutions, holding_time
                ),
                callback,
                with_token=True,
            )

    def _move_steps(self, steps, callback=None):
        """
        Callback method for open loop control.

        The amplitude of the movement is unitless, in steps. The n_factor and p_factor are user-friendly conversion
        factors corresponding to the estimated number of steps per degree or per millimeter (in the positive and
        negative directions).

        :param steps: number of steps to move, expressed in the unit of ATTR_POSITION
        :param callback: optional callback
        """
        # convert steps to the user-friendly unit deg or mm
        unit = self.get_value(ATTR_POSITION, UNIT)
        if unit in {"mdeg", "udeg", "ndeg", "pdeg"}:
            steps = kamzik3.units.Quantity(steps, unit).to("deg").m
        if unit in {"m", "um", "nm", "pm"}:
            steps = kamzik3.units.Quantity(steps, unit).to("mm").m

        n_factor, p_factor = self.get_value(ATTR_OPEN_LOOP_FACTOR)
        factor = n_factor if steps < 0 else p_factor
        steps = int(np.rint(steps * factor))
        if steps == 0:
            self.logger.info(
                f"Move by {steps} steps: minimum angular step = {1/factor:.3f} deg"
            )
        else:
            self.logger.info(f"Move by {steps} steps")
        return self.move_step(steps, callback=callback)

    def enable_close_loop(self):
        if self.position_attribute_copy is not None:
            self.set_attribute(
                (ATTR_POSITION, UNIT), self.position_attribute_copy[UNIT]
            )
            self.set_attribute(
                (ATTR_POSITION, OFFSET), self.position_attribute_copy[OFFSET]
            )
            self.set_attribute(
                (ATTR_POSITION, FACTOR), self.position_attribute_copy[FACTOR]
            )
            self[ATTR_POSITION].set_function = self._move_absolute
            self.position_attribute_copy = None

            # self.poll_command(":CHAN{}:POS:CURR?\r\n".format(self.channel), 250)
            # self.poll_command(":CHAN{}:POS:TARG?\r\n".format(self.channel), 250)

    def enable_open_loop(self):
        self.position_attribute_copy = self[ATTR_POSITION].attribute_copy()
        default_unit = self.get_value(ATTR_CLOSE_LOOP_UNIT)

        self.set_attribute(
            (ATTR_POSITION, UNIT),
            default_unit
            if default_unit is not None
            else self.position_attribute_copy[UNIT],
        )
        self.set_attribute((ATTR_POSITION, OFFSET), 0)
        self.set_attribute((ATTR_POSITION, FACTOR), 1)
        if self.get_value(ATTR_POSITION) is None:
            self.set_offsetted_value(ATTR_POSITION, 0)
        self[ATTR_POSITION].set_function = self._move_steps

    def set_move_method(self, value):
        if value == OPEN_LOOP:
            self.enable_open_loop()
        elif value == CLOSE_LOOP:
            self.enable_close_loop()

    @expose_method()
    def stop(self, callback=None):
        self.logger.info("Stop movement")
        return self.command(":S{}\n".format(self.channel), callback, with_token=True)

    @expose_method()
    def calibrate(self, callback=None):
        self.logger.info("Calibrate")
        return self.command(":CS{}\n".format(self.channel), callback, with_token=True)

    @expose_method()
    def find_reference(self, callback=None):
        self.logger.info("Reference")
        reference_direction = self.safe_direction_modes[
            self.get_value(ATTR_SAFE_DIRECTION)
        ]
        return self.command(
            ":FRM{},{},0,1\n".format(self.channel, reference_direction),
            callback,
            with_token=True,
        )

    @expose_method({"position": ATTR_POSITION})
    def move_absolute(self, position):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info("Move to absolute position {}".format(position))
            return self.set_value(ATTR_POSITION, position)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method({"position": ATTR_POSITION})
    def move_relative(self, position):
        try:
            position = device_units(self, ATTR_POSITION, position)
            self.logger.info("Move by relative position {}".format(position))
            return self.set_value(ATTR_POSITION, self[ATTR_POSITION].value() + position)
        except (UndefinedUnitError, ValueError) as e:
            self.logger.exception(e)
            raise DeviceUnitError(e)

    @expose_method(
        {
            "steps": "Number of steps",
            "amplitude": ATTR_STEP_AMPLITUDE,
            "frequency": ATTR_STEP_FREQUENCY,
        }
    )
    def move_step(self, steps, amplitude=None, frequency=None, callback=None):
        if amplitude in (None, ""):
            amplitude = self[ATTR_STEP_AMPLITUDE][VALUE]
        else:
            amplitude = device_units(self, ATTR_STEP_AMPLITUDE, amplitude).m
        if frequency in (None, ""):
            frequency = self[ATTR_STEP_FREQUENCY][VALUE]
        else:
            frequency = device_units(self, ATTR_STEP_FREQUENCY, frequency).m
        amplitude = int(self[ATTR_STEP_AMPLITUDE].remove_offset_factor(amplitude))
        frequency = int(self[ATTR_STEP_FREQUENCY].remove_offset_factor(frequency))
        self.logger.info(
            "Move by {} steps with amplitude {} and frequency of {} Hz".format(
                int(steps), int(amplitude), int(frequency)
            )
        )
        return self.command(
            ":MST{},{},{},{}\n".format(self.channel, int(steps), amplitude, frequency),
            callback,
            with_token=True,
        )

    def get_sensor_type(self):
        self.command(":GST{}\n".format(self.channel))

    def get_position_limit_range(self, sensor_type=None):
        if sensor_type in self.linear_sensor_type:
            self.command(":GPL{}\n".format(self.channel))
        else:
            self.command(":GAL{}\n".format(self.channel))

    def get_velocity(self):
        self.command(":GCLS{}\n".format(self.channel))

    def get_acceleration(self):
        self.command(":GCLA{}\n".format(self.channel))

    def get_safe_direction(self):
        self.command(":GSD{}\n".format(self.channel))

    def get_referenced(self):
        self.command(":GPPK{}\n".format(self.channel))

    def set_sensor_type(self, value, callback=None):
        self.logger.info("Set sensor type to {}".format(value))
        if value != self.get_attribute([ATTR_SENSOR_TYPE, PREVIOUS]):
            self.remove_poll_command(":GP{}\n".format(self.channel), 250)
            self.remove_poll_command(":GA{}\n".format(self.channel), 250)
            print(":SST{},{}\n".format(self.channel, int(value)))
            token = self.command(
                ":SST{},{}\n".format(self.channel, int(value)),
                callback,
                with_token=True,
            )
            self.get_sensor_type()
            return token
        return None

    def set_lower_position_limit(self, value, callback=None):
        self.logger.info("Set lower position limit to {}".format(value))
        # upper_limit = self[ATTR_POSITION_UPPER_LIMIT][VALUE]
        upper_limit = self.get_raw_value(ATTR_POSITION_UPPER_LIMIT)
        # self.set_value(ATTR_POSITION_LOWER_LIMIT, int(value))
        if upper_limit is None:
            upper_limit = value
            # self.set_value(ATTR_POSITION_UPPER_LIMIT, upper_limit)
        if int(self[ATTR_SENSOR_TYPE][VALUE]) in self.linear_sensor_type:
            if value > upper_limit:
                value, upper_limit = upper_limit, value
            out = self.command(
                ":SPL{},{},{}\n".format(self.channel, int(value), int(upper_limit)),
                with_token=True,
            )
        else:
            revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
            min_revolutions = int(value // revolution)
            min_angular_position = int(value - min_revolutions * revolution)
            max_revolutions = int(upper_limit // revolution)
            max_angular_position = int(upper_limit - max_revolutions * revolution)
            out = self.command(
                ":SAL{},{},{},{},{}\n".format(
                    self.channel,
                    int(min_angular_position),
                    int(min_revolutions),
                    int(max_angular_position),
                    int(max_revolutions),
                ),
                callback,
                with_token=True,
            )

        self.set_value(ATTR_POSITION, value, key=MIN)
        self.get_position_limit_range()
        return out

    def set_upper_position_limit(self, value, callback=None):
        self.logger.info("Set upper position limit to {}".format(value))
        lower_limit = self.get_raw_value(
            ATTR_POSITION_LOWER_LIMIT
        )  # self[ATTR_POSITION_LOWER_LIMIT][VALUE]
        # self.set_value(ATTR_POSITION_UPPER_LIMIT, int(value))
        if lower_limit is None:
            lower_limit = value
            # self.set_value(ATTR_POSITION_LOWER_LIMIT, lower_limit)
        if int(self[ATTR_SENSOR_TYPE][VALUE]) in self.linear_sensor_type:
            if value < lower_limit:
                value, lower_limit = lower_limit, value
            out = self.command(
                ":SPL{},{},{}\n".format(self.channel, int(lower_limit), int(value)),
                callback,
                with_token=True,
            )
        else:
            revolution = 360 * 10 ** abs(self[ATTR_ENCODER_RESOLUTION][VALUE])
            max_revolutions = int(value // revolution)
            max_angular_position = int(value - max_revolutions * revolution)
            min_revolutions = int(lower_limit // revolution)
            min_angular_position = int(lower_limit - min_revolutions * revolution)
            out = self.command(
                ":SAL{},{},{},{},{}\n".format(
                    self.channel,
                    int(min_angular_position),
                    int(min_revolutions),
                    int(max_angular_position),
                    int(max_revolutions),
                ),
                callback,
                with_token=True,
            )

        self.set_value(ATTR_POSITION, value, key=MAX)
        self.get_position_limit_range()
        return out

    def set_step_frequency(self, value, callback=None):
        self.logger.info("Set step frequency to {} Hz".format(value))
        return self.command(
            ":SCLF{},{}\n".format(self.channel, int(value)), callback, with_token=True
        )

    def set_step_amplitude(self, value):
        self.logger.info("Set step amplitude to {}".format(value))
        self[ATTR_STEP_AMPLITUDE][VALUE] = value

    def set_velocity(self, value, callback=None):
        self.logger.info("Set velocity to {}".format(value))
        return self.command(
            ":SCLS{},{}\n".format(self.channel, int(value)), callback, with_token=True
        )

    def set_acceleration(self, value, callback=None):
        self.logger.info("Set acceleration to {}".format(value))
        return self.command(
            ":SCLA{},{}\n".format(self.channel, int(value)), callback, with_token=True
        )

    def set_safe_direction(self, value, callback=None):
        self.logger.info("Set safe direction to {}".format(value))
        value = self.safe_direction_modes[value]
        return self.command(
            ":SSD{},{}\n".format(self.channel, value), callback, with_token=True
        )



import math
from time import perf_counter_ns
from typing import Any, Callable, Dict, List, Optional, Tuple, Union, KeysView

import pint
from pint import Quantity

from numpy import (
    dtype,
    float16,
    float32,
    float64,
    uint16,
    uint32,
    uint64,
    int16,
    int32,
    int64,
    uint8,
    int8,
    ndarray,
)

from kamzik3 import units
from kamzik3.constants import *
from kamzik3.constants import SAVE_CHANGE
from kamzik3.snippets.snippetDataAccess import is_equal


class SetFunction:
    """
    Decorator for setting Attributes value.

    Used when attribute value is not set directly, but via set function.
    """

    def __init__(self, attribute, callback=None):
        """
        Initiate for which attribute we need to setup flag and callback.

        :param Attribute attribute: Desired Attribute
        :param Callable callback: Callback after value was set
        """
        self.attribute = attribute
        self.callback = callback

    def __enter__(self):
        """
        Set flag and callback for desired Attribute.

        :return Attribute
        """
        self.attribute.with_set_function = True
        self.attribute.set_callback = self.callback
        return self.attribute

    def __exit__(self, value_type, value, traceback):
        """
        Clear set_function and callback after value was set.

        :param value_type:
        :param value:
        :param traceback:
        """
        self.attribute.with_set_function = False
        self.attribute.set_callback = None


class Attribute(dict):
    """
    Class definition of Attribute.

    Attribute extends Dictionary with predefined and user defined keys.

    Predefined keys are:

        VALUE: value of the attribute (value of specified type)
        TYPE: type of the attribute (int, float, ...)
        UNIT: desired unit of Attributes values (mm, degC, degK, ...)
        READONLY: value is readonly for Client
        DESCRIPTION: short description of Attributes purpose / meaning
        SETPOINT: when using setFunction, SETPOINT is first set
        SAVE_CHANGES: when set to True, after value is set, it's also written and save
        in the config file
        ALLOW_BROADCAST: allow broadcasting changes of Attribute to Client
        PREVIOUS: previous VALUE of Attribute after new VALUE was set

    Numerical Type keys:

        OFFSET: value offset (value + offset)
        FACTOR: value factor (value * factor)
        DECIMALS: specify number of decimals after decimal point
        MIN: minimum allowed value
        MAX: maximum allowed value
        TOLERANCE: tuple of (negative, positive) tolerance.
                   It's use to determine if value is set within specified tolerances.
    """

    # Holds main attributes, which each dict has to contain in order to be convertible
    # into Attribute
    attribute_keys = {VALUE, TYPE, UNIT, READONLY, DESCRIPTION}
    integer_types = (int8, int16, int32, int64, uint8, uint16, uint32, uint64)
    float_types = (float16, float32, float64)
    unsigned_types = (uint8, uint16, uint32, uint64)
    # Helper flag for SetFunction decorator
    with_set_function = False
    # Holds last time, that Attribute value was streamed to clients
    last_broadcast_time = last_update_time = perf_counter_ns()
    # Holds diff between last_stream_time and current time
    last_update_time_diff = 0
    # Indicates if attribute is numerical. It's set automatically based on TYPE.
    numerical = False
    # Token identifier associate with setting of new value
    set_token = 0
    # Any callable function to call after value was set
    set_callback = None
    # Minimum time in between of value change stream.
    # Every Attribute change between 0 and min_broadcast_timeout would not be broadcast
    # to the client.
    # Set this attribute to zero to stream every change.
    # Default value is in nanoseconds (100ms)
    min_broadcast_timeout = 0 * 1e6

    def __init__(
        self,
        attribute_id=None,
        default_value=None,
        min_value=None,
        max_value=None,
        unit: Optional[str] = None,
        set_function=None,
        default_type="str",
        decimals=0,
        readonly=False,
        description=None,
        set_value_when_set_function=True,
        tolerance=(0, 0),
        factor=1,
        offset=0,
        save_change=False,
        allow_broadcast=True,
        **kwargs
    ):
        """
        Initiate new Attribute object.

        When using with Device, always use device.create_attribute method.

        :param str attribute_id: Attribute id, it must be unique in context of one
         Device
        :param mixed default_value: Default VALUE
        :param float min_value: MIN value, ignored for str TYPE.
         Means inf if None for Numerical value.
        :param float max_value: MAX value, ignored for str TYPE.
         Means inf if None for Numerical value.
        :param str unit: Main UNIT of the attribute
        :param Callable set_function: Function to call when setting the VALUE
        :param str default_type: Default TYPE
        :param uint decimals: Number of DECIMALS after decimal point
        :param bool readonly: Specify if Attribute is READONLY for client
        :param str description: Short DESCRIPTION
        :param bool set_value_when_set_function: Set VALUE at the same time as calling
         the set_function
        :param list tolerance: (negative tolerance, positive tolerance)
        :param float factor: FACTOR to multiply with VALUE
        :param float offset: OFFSET to add to VALUE
        :param bool save_change: Save VALUE into config file as well
        :param bool allow_broadcast: Allow broadcast VALUE change to the client
        :param dict kwargs: User defined keys
        """
        dict.__init__(self)
        self._callbacks: Dict[Callable, List] = {}
        self.attribute_id = attribute_id
        self.set_function = set_function
        self.default_type = default_type

        self[OFFSET] = offset
        self[FACTOR] = factor
        # Set attribute value flag when using set function.
        # This flag prevents change of Attribute's value before value is actually
        # set on device side.
        self.set_value_when_set_function = set_value_when_set_function
        if unit is None:
            unit = ""
        self[UNIT] = unit
        self.min_broadcast_timeout = kwargs.get("min_broadcast_timeout", 0) * 1e6
        default_type = kwargs.get(TYPE, default_type)
        if default_type == TYPE_LIST:
            # Special type LIST
            self[TYPE] = TYPE_LIST
        elif isinstance(default_type, (list, KeysView)):
            #  Used for Combox implementation
            self[TYPE] = TYPE_LIST
            self[TYPE_LIST_VALUES] = list(default_type)
        elif default_type == TYPE_ARRAY or isinstance(default_type, ndarray):
            # Special type Array
            self[TYPE] = TYPE_ARRAY
        else:
            # Default TYPE is either string or Numerical
            default_type = dtype(default_type)
            self[TYPE] = dtype(default_type).name
            if default_type in self.integer_types:
                self[DECIMALS] = 0
                if min_value is None and default_type in self.unsigned_types:
                    self[MIN] = 0
                else:
                    self[MIN] = int(min_value) if min_value is not None else -math.inf
                self[MAX] = int(max_value) if max_value is not None else math.inf
                self.numerical = True
            elif default_type in self.float_types:
                self[DECIMALS] = decimals
                self[MIN] = float(min_value) if min_value is not None else -math.inf
                self[MAX] = float(max_value) if max_value is not None else math.inf
                self.numerical = True
        if self.numerical:
            self[TOLERANCE] = [float(tolerance[0]), float(tolerance[1])]
        self[READONLY] = readonly
        self[DESCRIPTION] = description
        self[VALUE] = default_value
        self[SETPOINT] = None
        self[PREVIOUS] = None
        self[SAVE_CHANGE] = save_change
        self[ALLOW_BROADCAST] = allow_broadcast

        # Update Dict with user defined key -> value pairs
        self.update(kwargs)

    def _update_set_time(self):
        """Set last_update_time and last_update_time_diff for future reference."""
        update_time = perf_counter_ns()
        self.last_update_time_diff = update_time - self.last_update_time
        self.last_update_time = update_time

    def allow_broadcast(self) -> bool:
        """
        Allow broadcast VALUE change.

        if ALLOW_BROADCASt is True and broadcast_timeout was exceeded

        :return: bool
        """
        return (
            self[ALLOW_BROADCAST]
            and (perf_counter_ns() - self.last_broadcast_time)
            >= self.min_broadcast_timeout
        )

    def __setitem__(self, k, v):
        """
        Called when value of the key is changed.

        :param str k: dict key
        :param mixed v: key value
        """
        if k == VALUE:
            self._update_set_time()
            try:
                self[PREVIOUS] = self[VALUE]
            except KeyError:
                pass
            if isinstance(v, Quantity):
                try:
                    v = v.to(self[UNIT]).m
                except AttributeError:
                    v = v.m
        elif k == TOLERANCE:
            # Make sure, that tolerance is of float type
            v = [float(t) for t in v]
        elif k == UNIT and v is not None:
            if v == "%":
                # We have to treat percent sign and turn in into 'percent' value
                v = "percent"
            if v not in ("", None):
                # Unit does not exists in pint units dictionary
                if units.parse_unit_name(v) == ():
                    # We define that new unit quantity equals 1
                    units.define("{}=1".format(v))
                    units.derived_units.append(v)
        try:
            # Return, if current and new value is equal
            if is_equal(self[k], v):
                return
        except KeyError:
            pass

        # We want to set VALUE using set_function
        if k == VALUE and self.set_function and self.with_set_function:
            self[SETPOINT] = v
            setpoint = self.remove_offset_factor(v)
            # Release set function flag
            self.with_set_function = False
            if self.set_value_when_set_function:
                # Store actual attribute[k] value into dictionary
                super().__setitem__(k, v)
            # Call set function
            if hasattr(self.set_function, "callback"):
                self.set_token = self.set_function(setpoint, callback=self.set_callback)
            else:
                # We don't have callback in a function, so just call it
                self.set_token = self.set_function(setpoint)
            if not self.set_value_when_set_function:
                return
        else:
            # Store actual attribute[k] value into dictionary
            super().__setitem__(k, v)

        self.handle_callbacks(k, v)

    def handle_callbacks(self, k, v):
        """
        Here comes a short description of handle_callbacks.

        When any parameter of attribute is set, call all relevant callbacks that are
        attached to the parameter.

        :param k: attribute key
        :type k: hashable
        :param v: attribute value
        :type v: object
        """
        # Call all relevant callbacks
        for (
            callback,
            (
                max_update_rate,
                updated_at,
                key_filter,
            ),
        ) in self._callbacks.copy().items():
            # Check update rate, continue if it's less than max_update_rate
            if (
                k == VALUE
                and max_update_rate is not None
                and (self.last_update_time - updated_at) <= max_update_rate
            ):
                continue
            # Check if key_filter is set
            if key_filter is None:
                # Key can be any value, therefore callback with key and value
                callback(k, v)
            elif k == key_filter:
                # Since we have filter set, callback with just value
                callback(v)
            else:
                continue
            try:
                # Set last callback time
                self._callbacks[callback][1] = perf_counter_ns()
            except KeyError:
                # Callback no longer exists, skipping
                pass

    def attribute_copy(self):
        """
        Create copy of Attribute.

        :return Attribute: Copy of this Attribute
        """
        return Attribute(**self)

    @staticmethod
    def from_dict(dictionary: Union[Dict[str, Any], Any]) -> Union["Attribute", Any]:
        """
        Create Attribute from input dictionary (recursively).

        Check if input dictionary can be converted into Attribute.

        :param dict dictionary: Input dictionary
        :return Attribute: Attribute if convertible
        """
        if isinstance(dictionary, dict):
            if Attribute.is_attribute(dictionary):
                # Check if unit is valid, if not add it's definition to Pint
                if dictionary[UNIT] not in ("", None):
                    try:
                        units.get_name(dictionary[UNIT])
                    except pint.errors.UndefinedUnitError:
                        units.define("{}=1".format(dictionary[UNIT]))
                        units.derived_units.append(dictionary[UNIT])
                return Attribute(**dictionary)
            else:
                for key, attribute in dictionary.items():
                    dictionary[key] = Attribute.from_dict(attribute)
            return dictionary
        else:
            return dictionary

    @staticmethod
    def is_attribute(dictionary) -> bool:
        """
        Check if provided dictionary is convertible to Attribute.

        Attribute.attribute_keys must be subset of Input dictionary keys.

        :param dict dictionary: Input dictionary
        :return: bool
        """
        try:
            return Attribute.attribute_keys.issubset(dictionary.keys())
        except (AttributeError, TypeError, ValueError):
            return False

    def attach_callback(
        self,
        callback: Callable,
        max_update_rate: Optional[float] = None,
        key_filter: Optional[str] = None,
    ):
        """
        Attach callback to Attribute.

        :param Callable callback: callable function
        :param float max_update_rate: max update rate in ms
        :param str key_filter:
        """
        assert callable(callback)

        if callback not in self._callbacks:
            if max_update_rate is not None:
                # Convert rate to nanoseconds
                max_update_rate *= 1e6
            self._callbacks[callback] = [max_update_rate, perf_counter_ns(), key_filter]

    def detach_callback(self, callback):
        """
        Detach callback from attribute.

        :param Callable callback: callable function
        """
        assert callable(callback)
        try:
            del self._callbacks[callback]
        except KeyError:
            pass  # Callback no longer exists

    def apply_offset_factor(self, value=None):
        """
        Apply offset and factor to attribute value.

        :param value: numerical value
        :return: mixed
        """
        if value is None:
            value = self[VALUE]

        if value is not None and self.numerical:
            return (value + self[OFFSET]) * self[FACTOR]
        elif self.default_type is bool:
            return self[FACTOR] == value
        else:
            return value

    def remove_offset_factor(self, value=None):
        """
        Remove offset and factor from attribute value.

        :param value: numerical value
        :return: mixed
        """
        if value is None:
            value = self[VALUE]
        if value is not None and self.numerical:
            return (value / self[FACTOR]) - self[OFFSET]
        elif self.default_type is bool:
            return self[FACTOR] == value
        else:
            return value

    def read_and_reset_token(self) -> int:
        """
        Return token and reset it back to 0.

        :return: int
        """
        token = self.set_token
        self.set_token = 0
        return token

    @staticmethod
    def list_attribute(attribute: Union[str, List, Tuple]) -> List:
        """
        Create list from input attribute.

        :param attribute: mixed
        :return: list
        """
        if isinstance(attribute, list):
            return attribute
        elif isinstance(attribute, tuple):
            return list(attribute)
        else:
            return [attribute]

    def convert_units(self, value: Optional[Quantity]):
        """
        Convert input value to current attribute unit.

        :param Quantity value: Quantity
        :return: value or Quantity
        """
        assert isinstance(value, Quantity)
        if value.dimensionless:
            if self[UNIT] is None:
                return value
            else:
                return units.Quantity("{} {}".format(value.m, self[UNIT]))

        return value.to(self[UNIT])

    def within_limits(self, value: Optional[Quantity]) -> bool:
        """
        Check if value is within limits.

        MIN <= VALUE <= MAX

        :param Quantity value:  Quantity
        :return: bool
        """
        assert isinstance(value, Quantity)
        return bool(self.minimum() <= value <= self.maximum())

    def within_tolerance(self, value: Optional[Quantity]) -> bool:
        """
        Check if value is within tolerance.

        :param Quantity value:  Value to check
        :return: bool
        """
        assert isinstance(value, Quantity)
        setpoint = self.setpoint()
        return bool(
            (setpoint - self.negative_tolerance())
            <= value
            <= (setpoint + self.positive_tolerance())
        )

    def value(self):
        """
        Get Attribute's value.

        :return: a Quantity or the non-numerical value of the attribute
        """
        if self.numerical:
            try:
                return units.Quantity(self[VALUE], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[VALUE])
            except TypeError:
                return self[VALUE]
        else:
            return self[VALUE]

    def minimum(self):
        """
        Get Attribute's minimum value.

        :return: a Quantity or the non-numerical value of the attribute's minimum
        """
        if self.numerical:
            try:
                return units.Quantity(self[MIN], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[MIN])
            except TypeError:
                return self[MIN]
        else:
            return self[MIN]

    def maximum(self):
        """
        Get Attribute's maximum value.

        :return: a Quantity or the non-numerical value of the attribute's maximum
        """
        if self.numerical:
            try:
                return units.Quantity(self[MAX], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[MAX])
            except TypeError:
                return self[MAX]
        else:
            return self[MAX]

    def negative_tolerance(self):
        """
        Get Attribute's negative tolerance.

        :return: a Quantity or the non-numerical value of the attribute's negative
         tolerance
        """
        if self.numerical:
            try:
                return units.Quantity(self[TOLERANCE][0], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[TOLERANCE][0])
            except TypeError:
                return self[TOLERANCE][0]
        else:
            return self[TOLERANCE][0]

    def positive_tolerance(self):
        """
        Get Attribute's positive tolerance.

        :return: a Quantity or the non-numerical value of the attribute's positive
         tolerance
        """
        if self.numerical:
            try:
                return units.Quantity(self[TOLERANCE][1], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[TOLERANCE][1])
            except TypeError:
                return self[TOLERANCE][1]
        else:
            return self[TOLERANCE][1]

    def offset(self):
        """
        Get Attribute's value offset.

        :return: a Quantity or the non-numerical value of the attribute's offset
        """
        if self.numerical:
            try:
                return units.Quantity(self[OFFSET], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[OFFSET])
            except TypeError:
                return self[OFFSET]
        else:
            return self[OFFSET]

    def setpoint(self):
        """
        Get Attribute's setpoint.

        :return: a Quantity or the non-numerical value of the attribute's setpoint
        """
        if self.numerical:
            try:
                return units.Quantity(self[SETPOINT], self[UNIT])
            except (pint.UndefinedUnitError, AttributeError):
                return units.Quantity(self[SETPOINT])
            except TypeError:
                return self[SETPOINT]
        else:
            return self[SETPOINT]

    def unit(self) -> Any:
        """
        Get Attribute's unit.

        :return: str
        """
        return self[UNIT]



import numpy as np
from typing import List

from kamzik3 import MacroException
from kamzik3 import session
from kamzik3.constants import *
from kamzik3.devices.general.deviceScanner import DeviceScanner
from kamzik3.macro.macro import Macro
from kamzik3.macro.step import StepDeviceAttributeNumerical
from kamzik3.snippets.snippetsDecorators import expose_method


class GeneralFlyScanner(DeviceScanner):
    attribute_filter: List[str] = [ATTR_VELOCITY, ATTR_POSITION]

    def _init_attributes(self):
        DeviceScanner._init_attributes(self)
        self.create_attribute(
            ATTR_DWELL_TIME,
            default_value=100,
            unit="ms",
            min_value=0,
            default_type=np.int64,
        )

    @expose_method()
    def get_scanner_attributes(self):
        return [ATTR_DWELL_TIME]

    @expose_method()
    def get_scanner_macro(self, scanner_input, scanner_attributes, parent_macro=None):
        device_id = scanner_input.step_attributes["device_id"]
        device = session.get_device(device_id)
        macro = Macro(device_id="{}_{}".format(self.device_id, scanner_input.common_id))
        distance = abs(scanner_input.start_point - scanner_input.end_point)
        dwell_time = scanner_attributes[ATTR_DWELL_TIME]
        velocity = device.get_attribute(ATTR_VELOCITY)
        maximum_velocity = velocity.maximum()
        scan_line_time = dwell_time * scanner_input.steps_count
        final_velocity = distance / scan_line_time
        final_velocity = final_velocity.to(maximum_velocity.u)

        if final_velocity > maximum_velocity:
            raise MacroException(
                "Calculated velocity {} is higher than maximum velocity {}.".format(
                    final_velocity, maximum_velocity
                )
            )

        scanner_input.set_steps_count(1)
        max_velocity_step = StepDeviceAttributeNumerical(
            "SetMaxVelocityStep",
            device_id,
            ATTR_VELOCITY,
            maximum_velocity,
            trigger_log=False,
            negative_tolerance=velocity.negative_tolerance(),
            positive_tolerance=velocity.positive_tolerance(),
            on_warning=scanner_input.step_attributes["on_warning"],
            timeout=scanner_input.step_attributes["timeout"],
        )
        if not scanner_input.bidirectional:
            macro.add(max_velocity_step)
            macro.add(
                StepDeviceAttributeNumerical(
                    "SetInitPositionStep",
                    device_id,
                    ATTR_POSITION,
                    scanner_input.start_point,
                    trigger_log=False,
                    negative_tolerance=scanner_input.step_attributes[
                        "negative_tolerance"
                    ],
                    positive_tolerance=scanner_input.step_attributes[
                        "positive_tolerance"
                    ],
                    on_warning=scanner_input.step_attributes["on_warning"],
                    timeout=scanner_input.step_attributes["timeout"],
                )
            )

        macro.add(
            StepDeviceAttributeNumerical(
                "FinalVelocityStep",
                device_id,
                ATTR_VELOCITY,
                final_velocity,
                trigger_log=False,
                negative_tolerance=velocity.negative_tolerance(),
                positive_tolerance=velocity.positive_tolerance(),
                on_warning=scanner_input.step_attributes["on_warning"],
                timeout=scanner_input.step_attributes["timeout"],
            )
        )

        previous_reset_step = scanner_input.get_reset_step()

        def get_reset_step():
            return [max_velocity_step, previous_reset_step]

        scanner_input.get_reset_step = get_reset_step
        macro.add(scanner_input)
        return macro


class GeneralFlyScannerMaxVelocity(DeviceScanner):
    attribute_filter = [ATTR_MAXIMUM_VELOCITY, ATTR_POSITION]

    def _init_attributes(self):
        DeviceScanner._init_attributes(self)
        self.create_attribute(
            ATTR_DWELL_TIME,
            default_value=100,
            unit="ms",
            min_value=0,
            default_type=np.int64,
        )

    @expose_method()
    def get_scanner_attributes(self):
        return [ATTR_DWELL_TIME]

    @expose_method()
    def get_scanner_macro(self, scanner_input, scanner_attributes, parent_macro=None):
        device_id = scanner_input.step_attributes["device_id"]
        device = session.get_device(device_id)
        macro = Macro(device_id="{}_{}".format(self.device_id, scanner_input.common_id))
        distance = abs(scanner_input.start_point - scanner_input.end_point)
        dwell_time = scanner_attributes[ATTR_DWELL_TIME]
        velocity = device.get_attribute(ATTR_MAXIMUM_VELOCITY)
        maximum_velocity = velocity.maximum()
        scan_line_time = dwell_time * scanner_input.steps_count
        final_velocity = distance / scan_line_time
        final_velocity = final_velocity.to(maximum_velocity.u)

        if final_velocity > maximum_velocity:
            raise MacroException(
                "Calculated velocity {} is higher than maximum velocity {}.".format(
                    final_velocity, maximum_velocity
                )
            )

        scanner_input.set_steps_count(1)
        max_velocity_step = StepDeviceAttributeNumerical(
            "SetMaxVelocityStep",
            device_id,
            ATTR_MAXIMUM_VELOCITY,
            maximum_velocity,
            trigger_log=False,
            negative_tolerance=velocity.negative_tolerance(),
            positive_tolerance=velocity.positive_tolerance(),
            on_warning=scanner_input.step_attributes["on_warning"],
            timeout=scanner_input.step_attributes["timeout"],
        )
        if not scanner_input.bidirectional:
            macro.add(max_velocity_step)
            macro.add(
                StepDeviceAttributeNumerical(
                    "SetInitPositionStep",
                    device_id,
                    ATTR_POSITION,
                    scanner_input.start_point,
                    trigger_log=False,
                    negative_tolerance=scanner_input.step_attributes[
                        "negative_tolerance"
                    ],
                    positive_tolerance=scanner_input.step_attributes[
                        "positive_tolerance"
                    ],
                    on_warning=scanner_input.step_attributes["on_warning"],
                    timeout=scanner_input.step_attributes["timeout"],
                )
            )

        macro.add(
            StepDeviceAttributeNumerical(
                "FinalVelocityStep",
                device_id,
                ATTR_MAXIMUM_VELOCITY,
                final_velocity,
                trigger_log=False,
                negative_tolerance=velocity.negative_tolerance(),
                positive_tolerance=velocity.positive_tolerance(),
                on_warning=scanner_input.step_attributes["on_warning"],
                timeout=scanner_input.step_attributes["timeout"],
            )
        )

        previous_reset_step = scanner_input.get_reset_step()

        def get_reset_step():
            return [max_velocity_step, previous_reset_step]

        scanner_input.get_reset_step = get_reset_step
        macro.add(scanner_input)
        return macro



import asynchat
import socket
from threading import RLock
from typing import Optional

from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsControlLoops import control_asyncore_loop


class DeviceSocket(asynchat.async_chat, Device):
    terminator: Optional[bytes] = b"\r\n"
    ac_in_buffer_size = 4096
    ac_out_buffer_size = 4096
    host = None
    port = None

    def __init__(self, host, port, device_id=None, config=None):
        Device.__init__(self, device_id, config)
        self.asyncore_loop = control_asyncore_loop
        self.host = host
        self.port = port
        self.send_lock = RLock()
        asynchat.async_chat.__init__(self)
        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.setblocking(False)
        self.buffer = []
        self.connect((host, port))

    def connect(self, address):
        try:
            asynchat.async_chat.connect(self, address)
            self.device_connection_poller.add_connecting_device(self)
        except socket.error:
            # Socket could not be connected. Non fatal error.
            # We don't want to reconnect immediately but rather wait for connection timeout.
            # Set devices as connecting, add it to device poller and wait until connection timeout.
            self.logger.exception("Socket creation exception")
            self.connecting = True
            self.device_connection_poller.add_connecting_device(self)
            # FIXME: Why is this a duplicate? Is socket.error the same as OSError?
        except (OverflowError, TypeError, OSError):
            self.logger.exception("Socket creation exception")
            self.close()

    def initiate_send(self):
        try:
            with self.send_lock:
                asynchat.async_chat.initiate_send(self)
        except IndexError:
            # Sending queue is already empty
            # Ignore this error
            self.logger.error("Send buffer is already empty")

    def collect_incoming_data(self, data: bytes) -> None:
        self.buffer.append(data.decode("utf-8"))

    def found_terminator(self):
        self.handle_readout(self.buffer)
        self.buffer = []

    def handle_connect(self):
        Device.handle_connect(self)

    def handle_connect_event(self):
        try:
            asynchat.async_chat.handle_connect_event(self)
            self.handle_configuration_event()
        except socket.error:
            self.logger.exception("Socket creation exception")
            self.close()

    def close_connection(self):
        self.stop_polling()
        return self.close()

    def close(self):
        try:
            self.socket.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass
        try:
            asynchat.async_chat.close(self)
        except OSError:
            pass
        self.set_status(STATUS_DISCONNECTED)

    def handle_expt_event(self):
        self.logger.error("Socket error")
        self.connection_error = True
        asynchat.async_chat.handle_expt_event(self)

    def handle_error(self):
        self.logger.exception("Uncaught exception")
        self.handle_response_error()

    def handle_configuration(self):
        raise NotImplementedError("Must be implemented in subclass")



import json
import logging
import os
import re
import time
from math import ceil
from threading import Lock
from typing import Any, Dict

import pandas as pd
from PyQt5.QtCore import (
    QAbstractTableModel,
    Qt,
    pyqtSlot,
    QModelIndex,
    pyqtSignal,
    QItemSelection,
)
from PyQt5.QtGui import QBrush, QColor, QKeySequence
from PyQt5.QtWidgets import (
    QWidget,
    QColorDialog,
    QHeaderView,
    QShortcut,
    QButtonGroup,
    QPushButton,
    QSpacerItem,
    QSizePolicy,
)
from natsort import humansorted

import kamzik3
from kamzik3 import DeviceError, SEP, session, units
from kamzik3.constants import *
from kamzik3.devices.deviceSavedAttributes import (
    UPDATE_INSERT_GROUPS,
    UPDATE_REMOVE_GROUP,
    UPDATE_INSERT_ROW,
    UPDATE_VALUE,
    UPDATE_REMOVE_ROW,
)
from kamzik3.gui.deviceWidget import DeviceWidget
from kamzik3.gui.templates.savedAttributesTemplate import Ui_Form
from kamzik3.snippets.snippetsUnits import device_units
from kamzik3.snippets.snippetsWidgets import (
    show_question_dialog,
    show_prompt_dialog,
    show_error_message,
)


class ViewModel(QAbstractTableModel):
    sig_value_update = pyqtSignal(int, int, str)
    colors: Dict[Any, Any] = {}

    def __init__(self, model_data, from_index=0, to_index=0, parent=None):
        self.model_data = model_data
        self.from_index = from_index
        self.to_index = to_index
        self.previous_table = None
        QAbstractTableModel.__init__(self, parent)

    # pylint: disable=unused-argument
    def columnCount(self, parent=None):
        columns = self.model_data.columns
        return len(columns) - 1

    # pylint: disable=unused-argument
    def rowCount(self, parent=None):
        display = self.to_index - self.from_index
        return display

    def data(self, index, role=None):
        row, column = index.row(), index.column()
        row += self.from_index + 1
        if role == Qt.DisplayRole:
            if column <= 1:
                return "   " + self.model_data.loc[row][column + 1] + "   "
            else:
                return self.model_data.loc[row][column + 1]
        elif role == Qt.BackgroundRole:
            color = self.model_data.loc[row][0]
            if color not in self.colors:
                self.colors[color] = QBrush(QColor(color))
            return self.colors[color]
        elif role == Qt.EditRole:
            current_value = self.model_data.loc[row][column + 1]
            return current_value
        return None

    def headerData(self, p_int, Qt_Orientation, role=None):
        if role == Qt.DisplayRole and Qt_Orientation == Qt.Horizontal:
            data = self.model_data.columns[p_int + 1]
            return data
        elif role == Qt.DisplayRole and Qt_Orientation == Qt.Vertical:
            return " " + str(self.from_index + p_int + 1) + " "
        return None

    def setData(self, index, value, role=None):
        if role == Qt.EditRole:
            row, column = index.row(), index.column()
            self.sig_value_update.emit(self.from_index + row + 1, column + 1, value)
            return True
        return False

    def flags(self, index):
        if index.column() == 1:
            return Qt.ItemIsSelectable | Qt.ItemIsEditable | Qt.ItemIsEnabled

        return Qt.ItemIsSelectable | Qt.ItemIsEnabled


class SavedAttributesWidget(Ui_Form, DeviceWidget):
    default_row_color = "#ffffff"
    file_path = None
    page_size = 100
    sig_refresh_group_tables = pyqtSignal(str)
    sig_refresh_group_table = pyqtSignal(bool, str, object)
    displayed_group_path = None
    column_pattern = re.compile(r"(.*) \[(.*)]\[(.*)]")

    def __init__(self, device=None, config=None, parent=None):
        self.rows_selection_mode = False
        self.button_group = QButtonGroup()
        self.update_lock = Lock()
        self.cache_h5_file = os.path.join(
            kamzik3.session.get_value(ATTR_CACHE_DIRECTORY), "client_save.h5"
        )
        self.cache_h5_content = None
        self.groups = {}
        self.logger = logging.getLogger("Gui.Device.SavedAttributeWidget")

        if device is None:
            # pylint: disable=non-parent-init-called
            QWidget.__init__(self, parent)
        else:
            DeviceWidget.__init__(self, device, config=config, parent=parent)

        self.manipulation_buttons = [
            self.button_set_value,
            self.button_change_color,
            self.button_remove,
        ]
        for button in self.manipulation_buttons:
            button.setDisabled(True)

        self.preset_group_buttons = (
            self.combo_preset,
            self.button_remove_preset_group,
            self.button_add_preset_group,
        )
        self.sig_refresh_group_tables.connect(self.slot_refresh_group_tables)
        self.sig_refresh_group_table.connect(self.slot_refresh_group_table)

        self.shortcut_x = QShortcut(QKeySequence("Alt+X"), self)
        self.shortcut_x.activated.connect(self.set_attribute_value)

        self.shortcut_a = QShortcut(QKeySequence("Alt+A"), self)
        self.shortcut_a.activated.connect(self.add_row)

        self.shortcut_d = QShortcut(QKeySequence("Alt+D"), self)
        self.shortcut_d.activated.connect(self.remove_rows)

        self.shortcut_c = QShortcut(QKeySequence("Alt+C"), self)
        self.shortcut_c.activated.connect(self.select_color)

        self.table_attributes.setSelectionBehavior(self.table_attributes.SelectItems)
        self.table_attributes.setSelectionMode(self.table_attributes.SingleSelection)
        self.table_attributes.horizontalHeader().setSectionResizeMode(
            QHeaderView.Stretch
        )
        self.table_attributes.horizontalHeader().setStretchLastSection(True)
        self.table_attributes.doubleClicked.connect(self.on_double_click)

    @pyqtSlot("PyQt_PyObject")
    def slot_set_status(self, value):
        if value in READY_DEVICE_STATUSES and not self.configured:
            self.configured = False
            self.slot_handle_configuration()
            self.device.attach_attribute_callback(
                ATTR_LAST_UPDATE, self.content_update, key_filter=VALUE
            )
            with open(self.cache_h5_file, "wb") as h5_fp:
                h5_fp.write(self.device.get_value(ATTR_CONTENT))
            self.device.attach_attribute_callback(
                ATTR_GROUPS, self.groups_changed, key_filter=VALUE
            )
            self.configured = True

    @pyqtSlot(object)
    def groups_changed(self, groups):
        if groups is None:
            return

        selected_item = self.combo_preset.currentText()
        self.combo_preset.clear()
        self.groups = {}
        # Add only top groups to selection
        for group in groups:
            group_path = group.split("/")[1:]
            if group_path[0] not in self.groups:
                self.groups[group_path[0]] = [group_path[1]]
            else:
                self.groups[group_path[0]].append(group_path[1])

        # We want alphabetically sorted group names
        sorted_groups = humansorted(self.groups.keys())
        self.combo_preset.addItems(sorted_groups)
        item_index = self.combo_preset.findText(selected_item)

        # Select group which was previously selected
        if item_index != -1:
            self.combo_preset.setCurrentIndex(item_index)
        else:
            self.combo_preset.setCurrentIndex(0)

    def slot_handle_configuration(self):
        pass

    @pyqtSlot()
    def content_update(self, value):
        if value is None or not self.configured:
            return
        with self.update_lock:
            values = value.split(SEP)
            if values[0] == UPDATE_INSERT_GROUPS:
                groups_data = json.loads(values[1])
                for group, data in groups_data.items():
                    df = pd.DataFrame(data, columns=list(data.keys()))
                    df.to_hdf(
                        self.cache_h5_file, key=group, format="table", complevel=1
                    )
            elif values[0] == UPDATE_REMOVE_GROUP:
                with pd.HDFStore(self.cache_h5_file) as hdf:
                    for group_path in self.device.get_value(ATTR_GROUPS):
                        if group_path.startswith(f"/{values[1]}/"):
                            hdf.remove(group_path)
            elif values[0] == UPDATE_INSERT_ROW:
                group = values[1]
                df = pd.read_hdf(self.cache_h5_file, key=group, mode="a")
                data = json.loads(values[2])
                position = int(values[3])
                line = pd.DataFrame(data, index=[position])
                df2 = pd.concat(
                    [df.iloc[: position - 1], line, df.iloc[position - 1 :]]
                )
                df2.reset_index(drop=True, inplace=True)
                df2.to_hdf(self.cache_h5_file, group, format="fixed")
                subgroup = values[1].split("/")
                if (
                    subgroup[1] == self.combo_preset.currentText()
                    and subgroup[2] == self.button_group.checkedButton().text()
                ):
                    self.sig_refresh_group_table.emit(True, subgroup[-1], df2)
            elif values[0] == UPDATE_VALUE:
                group = values[1]
                df = pd.read_hdf(self.cache_h5_file, key=group, mode="a")
                df.loc[int(values[2]), df.columns[int(values[3])]] = values[4]
                df.to_hdf(self.cache_h5_file, group, format="fixed")
                subgroup = values[1].split("/")
                if (
                    subgroup[1] == self.combo_preset.currentText()
                    and subgroup[2] == self.button_group.checkedButton().text()
                ):
                    self.sig_refresh_group_table.emit(True, subgroup[-1], df)
            elif values[0] == UPDATE_REMOVE_ROW:
                group = values[1]
                df = pd.read_hdf(self.cache_h5_file, key=group, mode="a")
                df.drop(labels=int(values[2]), axis=0, inplace=True)
                df.reset_index(drop=True, inplace=True)
                df.to_hdf(self.cache_h5_file, group, format="fixed")
                subgroup = values[1].split("/")
                if (
                    subgroup[1] == self.combo_preset.currentText()
                    and subgroup[2] == self.button_group.checkedButton().text()
                ):
                    self.sig_refresh_group_table.emit(True, subgroup[-1], df)

    @pyqtSlot(int, int, str)
    def slot_value_update(self, row, column, value):
        top_group = self.combo_preset.currentText()
        for sub_group in self.groups[top_group]:
            group_path = "/".join(("", top_group, sub_group))
            self.device.update_value(
                group=group_path, row=row, column=column, value=value
            )

    @pyqtSlot(QItemSelection, QItemSelection)
    # pylint: disable=unused-argument
    def current_selection_changed(self, selected, deselected):
        if self.rows_selection_mode:
            for button in self.manipulation_buttons + [self.button_add]:
                button.setDisabled(True)
        else:
            self.button_add.setDisabled(False)
            for button in self.manipulation_buttons:
                button.setDisabled(len(selected) == 0)

            if len(selected) > 0:
                if selected[0].topLeft().column() <= 1:
                    self.button_set_value.setDisabled(True)

    @pyqtSlot(QModelIndex)
    # pylint: disable=unused-argument
    def on_double_click(self, index):
        selected_index = self.table_attributes.selectedIndexes()[0]
        if self.checkbox_allow_double_click.isChecked() and selected_index.column() > 0:
            self.set_attribute_value()

    @pyqtSlot()
    def set_attribute_value(self):
        row, column = self.get_selected_cell()
        if column is None or column <= 1:
            return

        attribute_value = self.table_attributes.model().model_data.iloc[row, column]
        column_header = self.table_attributes.model().model_data.columns[column]
        result = self.column_pattern.search(column_header)
        if result is not None:
            try:
                device_id, attribute_name, unit = (
                    result.group(1),
                    result.group(2),
                    result.group(3),
                )
                self.logger.info(
                    f"Set {device_id} {attribute_name} to value {attribute_value} {unit}"
                )
                device = session.get_device(device_id)
                attribute_as_list = eval(f"[{attribute_name}]")
                value_in_device_units = device_units(
                    device, attribute_as_list, units.Quantity(attribute_value, unit)
                )
                device.set_value(attribute_as_list, value_in_device_units.m)
            except (DeviceError, Exception) as e:
                error_message = f"Cannot set {device_id} {attribute_name} to value {attribute_value} {unit}.\r\n{e}"
                show_error_message(error_message, self)

    def get_selected_cell(self):
        selected_index = self.table_attributes.selectedIndexes()
        if not selected_index:
            return None, None
        row = selected_index[0].row()

        if row is not None:
            row_offset = self.table_attributes.model().from_index
            row = row_offset + selected_index[0].row() + 1
            return row, selected_index[0].column() + 1
        else:
            return None, None

    @pyqtSlot()
    def remove_rows(self):
        with self.update_lock:
            row, column = self.get_selected_cell()
            if column is None:
                return

            answer = show_question_dialog(
                f"Do You really want to remove row {row} ?", "Confirmation", self
            )
            if answer:
                top_group = self.combo_preset.currentText()
                self.logger.info(f"Removing row {row} from group {top_group}")
                for sub_group in self.groups[top_group]:
                    group_path = "/".join(("", top_group, sub_group))
                    self.device.remove_row(group=group_path, row=row)

    @pyqtSlot()
    def add_row(self):
        with self.update_lock:
            user_comment = show_prompt_dialog("Type Your comment:", "Add new row", self)
            if user_comment:
                top_group = self.combo_preset.currentText()
                self.logger.info(f"Prepending new row in group {top_group}")
                for sub_group in self.groups[top_group]:
                    group_path = "/".join(("", top_group, sub_group))
                    df = pd.read_hdf(self.cache_h5_file, key=group_path)
                    table_row = [
                        "#fff",
                        time.strftime("%d-%m-%Y %H:%M:%S"),
                        user_comment,
                    ]
                    for column in df.columns[3:]:
                        result = self.column_pattern.search(column)
                        if result is not None:
                            try:
                                device_id, attribute_name, unit = (
                                    result.group(1),
                                    result.group(2),
                                    result.group(3),
                                )
                                attribute_as_list = eval(f"[{attribute_name}]")
                                device = session.get_device(device_id)
                                current_value = (
                                    device.get_attribute(attribute_as_list)
                                    .value()
                                    .to(unit)
                                )
                                table_row.append(current_value.m)
                            except (DeviceError, Exception):
                                table_row.append(None)
                        else:
                            table_row.append(None)
                    table_row = [str(v) for v in table_row]
                    self.device.insert_row(
                        group=group_path,
                        data=dict(zip(df.columns, table_row)),
                        position=2,
                    )

    @pyqtSlot()
    def select_color(self):
        with self.update_lock:
            row, _column = self.get_selected_cell()
            if row is None:
                return

            color = QColorDialog().getColor()
            if color is not None and color.name() != "#000000":
                self.slot_value_update(row, 0, str(color.name()))

    @pyqtSlot()
    def add_preset_group(self):
        with self.update_lock:
            group_name = show_prompt_dialog("Set name of a new group", "New save group")
            if group_name and group_name != "":
                self.logger.info(f"Adding new group {group_name}")
                groups = {}
                top_group = self.combo_preset.currentText()
                for sub_group in self.groups[top_group]:
                    group_path = "/".join(("", top_group, sub_group))
                    new_group_path = "/".join(("", group_name, sub_group))
                    df = pd.read_hdf(self.cache_h5_file, key=group_path)
                    groups[new_group_path] = dict(
                        zip(df.columns, [["None"]] * len(df.columns))
                    )
                self.device.insert_groups(groups_data=groups)

    @pyqtSlot()
    def remove_preset_group(self):
        with self.update_lock:
            current_group = self.combo_preset.currentText()
            answer = show_question_dialog(
                f"Do You really want to remove group {current_group} ?",
                "Confirmation",
                self,
            )
            if answer:
                self.logger.info(f"Removing group {current_group}")
                self.device.remove_group(group=current_group)

    @pyqtSlot("QString")
    def slot_refresh_group_tables(self, top_group):
        """
        Preset group was changed.

        We need to update attribute table.

        :param top_group: name of top group
        :type top_group: string
        """
        if top_group == "":
            return

        self.input_page.blockSignals(True)
        self.input_page.setValue(1)
        self.input_page.blockSignals(False)

        while self.layout_subgroups_select.count() != 0:
            self.layout_subgroups_select.removeItem(
                self.layout_subgroups_select.itemAt(0)
            )
        self.button_group = QButtonGroup()
        for button_index, subgroup in enumerate(self.groups[top_group]):
            button = QPushButton(subgroup)
            button.setStyleSheet(
                "QPushButton {background-color:silver;} QPushButton:checked {background-color:#f6b442;font-weight:bold}"
            )
            button.setCheckable(True)
            button.setAutoExclusive(True)
            button.setMinimumHeight(30)
            button.toggled.connect(
                lambda state, subgroup=subgroup: self.slot_refresh_group_table(
                    state, subgroup
                )
            )
            self.button_group.addButton(button, button_index)
            self.layout_subgroups_select.addWidget(button)
        self.layout_subgroups_select.addSpacerItem(
            QSpacerItem(0, 0, hPolicy=QSizePolicy.Expanding)
        )
        self.button_group.button(0).setChecked(True)
        self.check_page_borders()

    pyqtSignal(bool, str)
    pyqtSignal(bool, str, object)

    def slot_refresh_group_table(self, state, subgroup, model_data=None):
        if not state:
            return

        row_offset = (self.input_page.value() - 1) * self.page_size
        top_group = self.combo_preset.currentText()
        with self.update_lock:
            if model_data is None:
                model_data = pd.read_hdf(
                    self.cache_h5_file, "/".join([top_group, subgroup])
                )
            row_count = len(model_data.index) - 1
            to_index = (
                (row_offset + self.page_size)
                if row_count >= (row_offset + self.page_size)
                else row_count
            )
            view_model = ViewModel(model_data, from_index=row_offset, to_index=to_index)
            view_model.sig_value_update.connect(self.slot_value_update)
            self.table_attributes.setModel(view_model)
            self.table_attributes.horizontalHeader().setSectionResizeMode(
                0, QHeaderView.ResizeToContents
            )
            self.table_attributes.horizontalHeader().setSectionResizeMode(
                1, QHeaderView.Interactive
            )
            self.table_attributes.selectionModel().selectionChanged.connect(
                self.current_selection_changed
            )
            self.input_page.blockSignals(True)
            pages_count = ceil(row_count / self.page_size)
            self.input_page.setMaximum(pages_count if pages_count > 0 else 1)
            self.input_page.blockSignals(False)

    @pyqtSlot()
    def slot_jump_to_first_page(self):
        self.input_page.setValue(1)

    @pyqtSlot()
    def slot_jump_to_last_page(self):
        self.input_page.setValue(self.input_page.maximum())

    @pyqtSlot("int")
    def slot_jump_to_page(self, page_index):
        self.slot_refresh_group_table(True, self.button_group.checkedButton().text())
        self.check_page_borders()

    def check_page_borders(self):
        page_index = self.input_page.value()
        at_max = page_index == self.input_page.maximum()
        at_min = page_index == 1
        self.input_page.setSuffix(f" / {self.input_page.maximum()}")
        self.button_next_page.setDisabled(at_max)
        self.button_last_page.setDisabled(at_max)
        self.button_previous_page.setDisabled(at_min)
        self.button_first_page.setDisabled(at_min)



import re
import time
from typing import Optional

import numpy as np
import serial
from bidict import bidict

import kamzik3
from kamzik3.constants import *
from kamzik3.devices.devicePort import DevicePort
from kamzik3.devices.deviceSocket import DeviceSocket
from kamzik3.snippets.snippetsDecorators import expose_method

MEASUREMENT_SCALE = bidict(
    {
        "U": "Value in 1um",
        "H": "Value in 0.01mm",
        "Z": "Value in 0.1mm",
        "M": "Value in 1mm",
        "S": "Sensor unit",
        "R": "Raw value",
    }
)
MEASUREMENT_FACTORS = {
    "Value in 1um": 1000,
    "Value in 0.01mm": 100,
    "Value in 0.1mm": 10,
    "Value in 1mm": 1,
    "Sensor unit": 1,
    "Raw value": 1,
}
MEASUREMENT_UNITS = {
    "Value in 1um": "um",
    "Value in 0.01mm": "mm",
    "Value in 0.1mm": "mm",
    "Value in 1mm": "mm",
    "Sensor unit": "",
    "Raw value": "",
}


class DeviceBaumerLaserSensorPort(DeviceSocket):
    """
    Optoelectronic sensor, using laser to measure distance.

    Documentation at:
    https://svnsrv.desy.de/desy/cfel-cxi/doc/manuals/Baumer/OADM_13T7480_S35A_manual.pdf

    Example of config file:

    Laser_Sensor: &Laser_Sensor !Device:kamzik3.devices.deviceBaumerLaserSensor.DeviceBaumerLaserSensorSocket
    port: /dev/ttyS2
    device_id: Laser_Sensor
    config:
      attributes:
        !!python/tuple [Distance, Offset]: 50
        !!python/tuple [Measuring scale, Value]: Value in 0.1mm
    """

    error_pattern = re.compile(r"^{0E[A-Z0-9]+}$")
    terminator = b"}"
    push_commands_max = 1

    def handle_configuration(self):
        """Send Command '{0}' to get sensor configuration."""
        start_at = time.time()

        def _finish_configuration(*_, **__):
            self._config_commands()
            self._config_attributes()
            self.start_polling()
            self.set_status(STATUS_CONFIGURED)
            self.logger.info(
                "Device configuration took {} sec.".format(time.time() - start_at)
            )

        self.command("{0V}", callback=_finish_configuration)

    def _init_attributes(self):
        """
        Initialize the attributes.

        All values are readonly.

        Attributes:
            ATTR_SOFTWARE_VERSION: Version of sensor's SW
            ATTR_HARDWARE_VERSION: Version of sensor's HW
            ATTR_MANUFACTURE_DATE: Sensor's manufacture date
            ATTR_MODES: Modes of operation
            ATTR_MEASURING_SCALE: Measurement scale
            ATTR_DISTANCE: Measured distance
        """
        DeviceSocket._init_attributes(self)
        self.create_attribute(ATTR_SOFTWARE_VERSION, readonly=True)
        self.create_attribute(ATTR_HARDWARE_VERSION, readonly=True)
        self.create_attribute(ATTR_MANUFACTURE_DATE, readonly=True)
        self.create_attribute(ATTR_MODES, readonly=True)
        self.create_attribute(
            ATTR_MEASURING_SCALE,
            readonly=True,
            default_type=MEASUREMENT_SCALE.inv.keys(),
            set_function=self.set_measuring_scale,
        )
        self.create_attribute(
            ATTR_DISTANCE,
            default_value=0,
            default_type=np.float,
            readonly=True,
            unit="mm",
            decimals=4,
        )

    def handle_readout(self, readout_buffer):
        """
        Handle readout from Device.

        :param str readout_buffer: raw readout buffer from Device
        """
        command, output, callback, token = DevicePort.handle_readout(
            self, readout_buffer
        )
        is_error_message = self.error_pattern.match(output + "}")
        # checksum = output[-2:]  # to simplify communication, ignore checksum
        output = output[2:-2]
        if is_error_message is not None:
            # Handle device error
            self.handle_command_error(command, readout_buffer)
            command = RESPONSE_ERROR
        elif command == "{0V}":
            # Parse Device configuration
            # output_format = output[2]  # not used
            # waiting_time_pddm = output[3]  # not used
            readout_modes = output[18:]
            self.set_raw_value(ATTR_SOFTWARE_VERSION, output[4:10])
            self.set_raw_value(ATTR_HARDWARE_VERSION, output[10:12])
            self.set_raw_value(ATTR_MANUFACTURE_DATE, output[12:18])
            self.set_raw_value(ATTR_MODES, readout_modes)
            self.set_raw_value(ATTR_MEASURING_SCALE, MEASUREMENT_SCALE[output[1]])
            self.set_attribute(
                [ATTR_DISTANCE, UNIT], MEASUREMENT_UNITS[MEASUREMENT_SCALE[output[1]]]
            )
        elif command == "{0M}":
            # Parse and store measured distance
            values = re.match("^M([MA]+[0-9]+)+$", output)
            for group in values.groups():
                if group[0] == "M":
                    factor = MEASUREMENT_FACTORS[self.get_value(ATTR_MEASURING_SCALE)]
                    distance = int(group[1:]) / factor
                    self.set_raw_value(ATTR_DISTANCE, distance)

        if callback is not None:
            # Call Callback in case it's set
            self.handle_readout_callback(callback, command, readout_buffer)
        if token:
            # Publish command, readout_buffer in case token was set
            kamzik3.session.publisher.push_message(
                self.device_id, [command, str(readout_buffer)], token
            )

    def set_measuring_scale(self, value, callback=None) -> Optional[int]:
        """
        Set measuring scale.

        :param str value: One of MEASUREMENT_UNITS values
        :param Callable callback: Optional callback function
        :return int: token
        """
        self.logger.info("Setting measuring scale to {}".format(value))
        self.set_attribute([ATTR_DISTANCE, UNIT], MEASUREMENT_UNITS[value])
        return self.command(
            "{{0S{}}}".format(MEASUREMENT_SCALE.inverse.get(value)),
            callback,
            with_token=True,
        )

    @expose_method()
    def laser_on(self, callback=None) -> Optional[int]:
        """
        Turn on the laser.

        Measure distance continuously.
        :param Callable callback: Optional callback function
        :return int: token
        """
        self.logger.info("Setting laser On")
        return self.command("{0L1}", callback, with_token=True)

    @expose_method()
    def laser_off(self, callback=None) -> Optional[int]:
        """
        Torn off the laser.

        Stop measurement.
        :param Callable callback: Optional callback function
        :return int: token
        """
        self.logger.info("Setting laser Off")
        return self.command("{0L0}", callback, with_token=True)

    @expose_method()
    def measure_distance(self, callback=None) -> Optional[int]:
        """
        Sample out one distance measurement.

        Turn Laser on, Measure distance, Turn Lase off
        In our use case we don't need constant measurement.
        If You need in other use case constant measurement, just setup polling for
        '{0M}' command.

        :param Callable callback: Optional callback function
        :return int: token
        """
        self.logger.info("Measuring distance")
        self.laser_on()
        out = self.command("{0M}", callback, with_token=True)
        self.laser_off()
        return out

    def stop(self):
        """Turn off the laser."""
        self.logger.info("Stop device")
        self.laser_off()


# abstract methods `collect_incoming_data` and `found_terminator` are defined in
# DeviceSocket (base class of DeviceBaumerLaserSensorPort). Since multiple inheritance
# is used here we better disable the pylint warning.
# pylint: disable=abstract-method
# see https://github.com/python/mypy/issues/9319 for mypy error
class DeviceBaumerLaserSensorSocket(DevicePort, DeviceBaumerLaserSensorPort):  # type: ignore
    """Extend ETH controller to enable RS232 connection."""

    terminator = b"}"

    # pylint: disable=super-init-not-called
    def __init__(
        self,
        port,
        baud_rate=38400,
        parity=serial.PARITY_NONE,
        stop_bits=serial.STOPBITS_ONE,
        byte_size=serial.EIGHTBITS,
        device_id=None,
        config=None,
    ):
        DevicePort.__init__(
            self, port, baud_rate, parity, stop_bits, byte_size, device_id, config
        )

    def handshake(self) -> bool:
        """
        Confirm connection to Device via RS232.

        :return bool: True if handshake successful
        """
        self.logger.info("Handshake initiated")
        self.push(b"{0V}")
        time.sleep(0.05)  # Give some time for devices to reply
        if self.serial_port.in_waiting != 0:
            output = self.read_all().strip()
            is_error_message = self.error_pattern.match(output)
            return is_error_message is None
        return False



from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QFrame, QComboBox, QLabel, QHBoxLayout
from pint import UndefinedUnitError
from typing import List

from kamzik3 import units
from kamzik3.constants import UNIT
from kamzik3.snippets.snippetsUnits import get_attribute_unit_range
from kamzik3.snippets.snippetsWidgets import clear_layout


class AttributeUnitWidget(QFrame):
    current_unit = None
    sig_unit_changed = pyqtSignal("QString")
    combox = None

    def __init__(self, attribute, config=None, parent=None):
        self.attribute = attribute
        self.config = config
        if self.config is None:
            self.config = {}
        QFrame.__init__(self, parent=parent)
        self.setLayout(QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.set_unit(attribute[UNIT])

    def set_unit(self, unit):
        if self.combox is not None:
            try:
                self.combox.currentTextChanged.disconnect()
            except TypeError:
                pass
            self.combox = None
        clear_layout(self, self.layout())
        self.layout().addWidget(self.get_unit_widget(unit))
        # self.sig_unit_changed.emit(unit)

    def get_unit_widget(self, unit):
        if unit is not None:
            if unit in units.derived_units or self.config.get("static_unit", False):
                return self._get_static_unit_widget(unit)
            else:
                try:
                    return self._get_combox_unit_widget(unit)
                except (UndefinedUnitError, AttributeError, KeyError):
                    return self._get_static_unit_widget(unit)
        else:
            return QLabel("")

    def _get_static_unit_widget(self, unit):
        if unit == "percent":
            unit = "%"
        return QLabel(" {}".format(unit))

    def _get_combox_unit_widget(self, unit):
        self.combox = QComboBox()
        possible_units: List = get_attribute_unit_range(self.attribute)
        self.combox.addItems(possible_units)
        if unit in possible_units:
            index = possible_units.index(unit)
        else:
            # can happen when the limits are of smaller amplitude than 1 unit
            # then the unit does not correspond anymore to the attribute unit
            index = 0
        self.current_unit = possible_units[index]

        # Signal proxy
        self.combox.currentTextChanged.connect(self.sig_unit_changed.emit)
        self.combox.wheelEvent = lambda *args: None

        self.combox.setCurrentIndex(index)
        return self.combox

    def close(self):
        self.sig_unit_changed.disconnect()
        clear_layout(self, self.layout())
        self.attribute = None
        self.config = None



import datetime
import os
import re
import time
from threading import Thread

import numpy as np

from kamzik3 import DeviceError
from kamzik3.constants import *
from kamzik3.devices.device import Device
from kamzik3.snippets.snippetsDecorators import expose_method
from kamzik3.snippets.snippetsUnits import device_units

try:
    import PyTango
    from PyTango import DeviceProxy
except ImportError:
    raise DeviceError("PyTango import error!")

ATTR_TANGO_STATUS = "Tango status"

helptext = """
Example of yaml configuration

Lambda: !Device:kamzik3.devices.deviceTango.DeviceTango
    path: tango://mll-lambda.desy.de:10000/mll/lambda/01
    config:
      poll_attributes:
        - [SaveFileName, 300]
        - [FileStartNum, 300]
        - [SaveFilePath, 500]
        - [EnergyThreshold, 1000]
        - [ShutterTime, 1000]
        - [FilePreExt, 1000]
        - [FramesPerFile, 1000]
"""


class DeviceTango(Device, DeviceProxy):
    def __init__(self, path, device_id=None, config=None):
        self.path = path
        self.tango_attributes = []
        self.tango_commands = []
        self.tango_exceptions = (
            PyTango.ConnectionFailed,
            PyTango.CommunicationFailed,
            PyTango.DevFailed,
            PyTango.DeviceUnlocked,
        )
        Device.__init__(self, device_id, config)
        self.connect()

    def connect(self, *args):
        """
        Call only this function to connect devices to port / socket / library / ...
        :param args: connect attributes
        """
        self.connecting = True
        try:
            DeviceProxy.__init__(self, self.path)
        except self.tango_exceptions as e:
            self.handle_connection_error(e)
            return

        self.handle_configuration_event()
        self.connected = True
        self.connecting = False

    def handle_configuration(self):
        start_at = time.time()

        try:
            self.tango_attributes = set(self.get_attribute_list()) - set(
                self.attributes.keys()
            )
        except self.tango_exceptions as e:
            self.handle_connection_error(e)
            raise e

        self.tango_commands = self.command_list_query()

        for attribute in self.tango_attributes:
            try:
                ta = self.read_attribute(attribute)
            except self.tango_exceptions as e:
                self.handle_command_error(attribute, str(e))
                continue

            # FIXME: Can't we just disable the callback here? Check back when we have a type for the "set_function"
            # parameter for create_attribute
            def set_function(value, attribute_name=attribute, callback=None):
                try:
                    self.write_attribute(attribute_name, value)
                except (self.tango_exceptions, (AttributeError, TypeError)) as e:
                    raise DeviceError(e)

            tc = self.get_attribute_config(attribute)

            default_type = "str"
            value = ta.value
            min_value, max_value = None, None
            decimals = 0
            if isinstance(value, np.ndarray):
                continue
            if tc.data_type == PyTango._tango.CmdArgType.DevLong:
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = np.float32
                    min_value, max_value = self._get_tango_limits(tc, default_type)
                    value = float(value)
            elif tc.data_type == PyTango._tango.CmdArgType.DevBoolean:
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = bool
                    value = bool(value)
            elif tc.data_type in (
                PyTango._tango.CmdArgType.DevDouble,
                PyTango._tango.CmdArgType.DevFloat,
            ):
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = np.float64
                    min_value, max_value = self._get_tango_limits(tc, default_type)
                    value = float(value)
                    decimals_re = re.search(r"%[0-9]+\.([0-9]+)f", tc.format)
                    if decimals_re is not None:
                        decimals = int(decimals_re.group(1))
            elif tc.data_type == PyTango._tango.CmdArgType.DevULong:
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = np.uint32
                    min_value, max_value = self._get_tango_limits(tc, default_type)
                    value = int(value)
            elif tc.data_type == PyTango._tango.CmdArgType.DevShort:
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = np.int16
                    min_value, max_value = self._get_tango_limits(tc, default_type)
                    value = int(value)
            elif tc.data_type == PyTango._tango.CmdArgType.DevUShort:
                # Not array
                if tc.max_dim_x == 1 and tc.max_dim_y == 0:
                    default_type = np.uint16
                    min_value, max_value = self._get_tango_limits(tc, default_type)
                    value = int(value)
            else:
                value = str(value)

            self.create_attribute(
                attribute,
                default_value=value,
                readonly=not ta.nb_written,
                description=tc.description,
                default_type=default_type,
                set_function=set_function,
                unit=tc.unit,
                min_value=min_value,
                max_value=max_value,
                decimals=decimals,
            )

        for command in self.tango_commands:
            command_name = command.cmd_name
            # Since this device is already DeviceProxy all methods are callable
            # Just add them to exposed methods and we are done :)
            self.exposed_methods.append((command_name, None))

        self.start_polling()
        self._config_commands()
        self._config_attributes()
        self.set_status(STATUS_CONFIGURED)
        self.logger.info(
            "Device configuration took {} sec.".format(time.time() - start_at)
        )

    def start_polling(self):
        super().start_polling()
        self.poll_attribute("State", self.config.get("base polling rate", 100))
        self.poll_command("State", self.config.get("base polling rate", 100))

        for poll_attribute in self.config.get("poll_attributes", []):
            tango_attribute, timeout = poll_attribute
            self.poll_attribute(tango_attribute, timeout)
            self.poll_command(tango_attribute, timeout)

    def stop_polling(self):
        super().stop_polling()
        try:
            self.stop_poll_attribute("State")

            for poll_attribute in self.config.get("poll_attributes", []):
                tango_attribute, _timeout = poll_attribute
                self.stop_poll_attribute(tango_attribute)
        except self.tango_exceptions:
            pass

    def command(self, command, callback=None, with_token=False, returning=True):
        raise DeviceError("Device does not accept any commands.")

    def command_inout(self, param):
        try:
            return DeviceProxy.command_inout(self, param)
        except self.tango_exceptions as e:
            raise DeviceError(e)

    def send_command(self, commands):
        for command in commands:
            tango_attribute = command[0]
            try:
                polled_attribute = self.attribute_history(tango_attribute, 1)[0]
            except UnicodeDecodeError:
                continue
            except self.tango_exceptions as e:
                if e.args[0].reason == "API_CantConnectToDevice":
                    self.handle_connection_error(e)
                    return []
                elif e.args[0].reason == "API_NoDataYet":
                    return []
                else:
                    self.handle_command_error(command, e)
                    continue

            self.set_raw_value(tango_attribute, polled_attribute.value)
            if tango_attribute == "State":
                if polled_attribute.value in (
                    PyTango.DevState.MOVING,
                    PyTango.DevState.RUNNING,
                ):
                    self.set_status(STATUS_BUSY)
                elif polled_attribute.value in (
                    PyTango.DevState.ON,
                    PyTango.DevState.STANDBY,
                ):
                    self.set_status(STATUS_IDLE)
                elif polled_attribute.value == PyTango.DevState.INIT:
                    self.set_status(STATUS_CONFIGURING)
                elif polled_attribute.value in (
                    PyTango.DevState.DISABLE,
                    PyTango.DevState.UNKNOWN,
                ):
                    self.set_status(STATUS_DISCONNECTED)
                else:
                    self.set_status(STATUS_ERROR)
        self.response_timestamp = time.time()
        return []

    def valid_command_format(self, command):
        return True

    @staticmethod
    def _get_tango_limits(tc, type_cast):
        if tc.min_value == "Not specified":
            min_value = None
        else:
            min_value = type_cast(tc.min_value)
        if tc.max_value == "Not specified":
            max_value = None
        else:
            max_value = type_cast(tc.max_value)
        return min_value, max_value


class DeviceTangoLambdaDetector(DeviceTango):
    def handle_configuration(self):
        try:
            super().handle_configuration()
            current_mode = self.attributes["OperatingMode"][VALUE]
            self.create_attribute(
                "OperatingMode",
                group="Mode",
                default_value=current_mode,
                default_type=["ContinuousReadWrite", "TwentyFourBit"],
                set_function=self.set_mode,
            )
            self.attributes["FrameNumbers"][MIN] = 1
        except self.tango_exceptions:
            # Exception was handled on DeviceTango level
            pass

    @expose_method({"Mode": "Opertaing mode"})
    def set_mode(self, Mode):
        """
        For some reason changing mode is completely blocking action.

        There is no status indication, no documentation.
        We just have to stop polling and waiting for some time.
        To be safe waiting time is set to 10s.
        :param Mode:
        """
        if self.get_value("OperatingMode") != Mode:

            def _set():
                self.stop_polling()
                self.set_attribute([ATTR_ENABLED, VALUE], False)
                self.set_status(STATUS_BUSY)
                time.sleep(0.2)
                self.write_attribute_asynch("OperatingMode", Mode)
                self._set(["OperatingMode", VALUE], Mode)
                time.sleep(9)
                self.start_polling()
                self.set_attribute([ATTR_ENABLED, VALUE], True)

            Thread(target=_set).start()

    @expose_method(
        {
            "Name": "FilePrefix",
            "Exposure": "ShutterTime",
            "Frames": "FrameNumbers",
            "Save": "SaveAllImages",
        }
    )
    def acquire_frame(self, Name, Exposure, Frames, Save):
        now = datetime.datetime.now()
        self.set_attribute(["TriggerMode", VALUE], 0)
        self.set_attribute(["SaveFilePath", VALUE], self.config.get("shotDir"))
        self.set_attribute(
            ["FilePreExt", VALUE],
            "_{}_{}".format(self.device_id, now.strftime("%Y-%m-%d_%H:%M:%S")),
        )
        self.set_attribute(["FilePrefix", VALUE], Name)
        self.set_attribute(["FileStartNum", VALUE], 0)
        Exposure = device_units(self, "ShutterTime", Exposure).m
        self.set_attribute(["ShutterTime", VALUE], float(Exposure))
        Frames = device_units(self, "FrameNumbers", Frames).m
        self.set_attribute(["FrameNumbers", VALUE], int(float(Frames)))
        self.set_attribute(["SaveAllImages", VALUE], int(bool(Save)))
        self.StartAcq()

    @expose_method({"Exposure": "ShutterTime"})
    def live_view(self, Exposure):
        Exposure = device_units(self, "ShutterTime", Exposure).m
        self.set_attribute(["ShutterTime", VALUE], float(Exposure))
        self.set_attribute(["SaveAllImages", VALUE], False)
        self.set_attribute(["LiveMode", VALUE], True)
        self.set_attribute(["FrameNumbers", VALUE], 100000)
        self.set_attribute(["TriggerMode", VALUE], 0)
        self.StartAcq()

    @expose_method({"Exposure": "ShutterTime", "Frames": "FrameNumbers"})
    def acquire_scan_frame(self, Exposure, Frames):
        Exposure = float(device_units(self, "ShutterTime", Exposure).m)
        if Exposure != self.get_value("ShutterTime"):
            self.set_attribute(["ShutterTime", VALUE], Exposure)
        Frames = int(float(device_units(self, "FrameNumbers", Frames).m))
        if Frames != self.get_value("FrameNumbers"):
            self.set_attribute(["FrameNumbers", VALUE], Frames)
        time.sleep(1)
        self.StartAcq()

    @expose_method()
    def stop(self):
        self.StopAcq()


class DeviceTangoEigerDetector(DeviceTango):
    """
    Implementation of the Eiger detector.

    :param path: path of the Tango server, e.g.
     tango://haspp11oh:10000/p11/simplon_detector/eh.01
    :param file_writer: instance of a Simplon FileWriter Tango device
    :param eiger_stream: instance of a Simplon Stream Tango device
    :param device_id: str, the name to give to this device
    :param config: a dictionary of configuration parameters
    """

    def __init__(
        self, path, file_writer=None, eiger_stream=None, device_id=None, config=None
    ):
        self.file_writer = file_writer
        self.eiger_stream = eiger_stream
        DeviceTango.__init__(self, path, device_id, config)

    def handle_configuration(self):
        """Configure the Eiger device."""
        try:
            super().handle_configuration()
            self.check_interface()
            if self.file_writer is not None:
                self.share_group(self.file_writer, None, "FileWriter")
            if self.eiger_stream is not None:
                self.share_group(self.eiger_stream, None, "EigerStream")
        except self.tango_exceptions:
            # Exception was handled on DeviceTango level
            pass

    def check_interface(self):
        """
        Check the configuration of the Eiger interface.

        Either the FileWriter or the Streaming interface has to be enabled (not both).

        Mode 0=enabled, 1=Disabled
        """
        if (
            self.file_writer is not None
            and self.file_writer.get_value("Mode") == "0"
            and self.eiger_stream is not None
            and self.eiger_stream.get_value("Mode") == "0"
        ):
            raise DeviceError(
                "Enable either the File Writer or the Streaming Interface, not both"
            )

        if self.file_writer is None and self.eiger_stream is None:
            raise DeviceError("No interface provided for the Eiger!")

    def _init_attributes(self):
        DeviceTango._init_attributes(self)
        self.create_attribute("shotDir")
        self.create_attribute("scanDir")
        self.create_attribute("scanFilenamePattern")
        self.create_attribute(
            ATTR_FRAME_COUNT,
            default_value=0,
            readonly=True,
            description="Internal frame counter",
            default_type=np.uint64,
            min_value=0,
        )

    def _config_interface(self, saving_directory, prefix, nb_frames=1):
        """
        Configure the Eiger interface.

        :param saving_directory: str, the name of the saving directory
        :param prefix: str, prefix to add to the file names (filewriter) or folder name
         for the dataset (stream subsystem)
        :param nb_frames: int, number of frames to be saved per .h5 file for the
         Filewiter
        """
        now = datetime.datetime.now()
        if self.file_writer is not None:
            path = os.path.join(
                self.get_value(saving_directory),
                f"{prefix}_{self.device_id}_{now.strftime('%Y-%m-%d_%H:%M:%S')}",
            )
            self.set_attribute(["FileWriter", "Mode", VALUE], 0)
            self.set_attribute(["FileWriter", "NamePattern", VALUE], path)
            self.set_attribute(["FileWriter", "NimagesPerFile", VALUE], nb_frames)
        else:  # use the Stream subsystem
            path = '{"series_name": "' + prefix + '"}'
            self.set_attribute(["EigerStream", "Mode", VALUE], 0)
            self.set_attribute(["EigerStream", "HeaderAppendix", VALUE], path)
            self.set_attribute(["EigerStream", "ImageAppendix", VALUE], path)

    @expose_method({"Name": "FilePrefix", "Exposure": "FrameTime", "Frames": "Nimages"})
    def acquire_frame(self, Name, Exposure, Frames):
        Frames = int(device_units(self, "Nimages", Frames).m)
        self._config_interface(
            saving_directory="shotDir", prefix=Name, nb_frames=Frames
        )

        Exposure = float(device_units(self, "FrameTime", Exposure).m)
        # According to Jan 10ms is lowest frame time
        frame_time = 10e-3 if Exposure < 10e-3 else Exposure

        self.set_attribute(["TriggerMode", VALUE], 3)
        self.set_attribute(["FrameTime", VALUE], frame_time)
        self.set_attribute(["CountTime", VALUE], Exposure)
        self.set_attribute(["Ntrigger", VALUE], 1)
        self.set_attribute(["Nimages", VALUE], Frames)
        self.Arm()
        self.Trigger()
        self.set_value(ATTR_FRAME_COUNT, self.get_value(ATTR_FRAME_COUNT) + Frames)

    @expose_method({"Exposure": "FrameTime"})
    def live_view(self, Exposure):
        """
        Live view of the Eiger detector.

        This is useful when aligning the optics or the sample. Images will not be saved.

        :param Exposure: desired exposure time per frame
        """
        self.logger.info(f"Live view has started.")
        self.set_attribute(["TriggerMode", VALUE], 3)
        if self.file_writer is not None:
            self.logger.info(f"FileWriter disabled")
            self.set_attribute(["FileWriter", "Mode", VALUE], 1)
            self.set_attribute(["FileWriter", "NimagesPerFile", VALUE], 1000000)
        if self.eiger_stream is not None:
            self.logger.info(f"EigerStream disabled")
            self.set_attribute(["EigerStream", "Mode", VALUE], 1)
        Exposure = float(device_units(self, "FrameTime", Exposure).m)
        # According to Jan 10ms is lowest frame time
        frame_time = 10e-3 if Exposure < 10e-3 else Exposure

        self.set_attribute(["FrameTime", VALUE], frame_time)
        self.set_attribute(["CountTime", VALUE], Exposure)
        self.set_attribute(["Ntrigger", VALUE], 1)
        self.set_attribute(["Nimages", VALUE], 1000000)
        self.Arm()
        self.Trigger()
        self.set_value(ATTR_FRAME_COUNT, self.get_value(ATTR_FRAME_COUNT) + 1000000)

    @expose_method()
    def stop(self):
        if self.get_value(ATTR_STATUS) == STATUS_BUSY:
            self.logger.info("Acquisition aborted")
            self.Abort()
            self.Disarm()
